@@ -23,7 +23,7 @@
 #define require(expr)                                           \
   do {                                                          \
       if (!(expr)) {                                            \
-          PL_ERR(gpu, "Validation failed: %s (%s:%d)\n",        \
+          PL_ERR(gpu, "Validation failed: %s (%s:%d)",          \
                   #expr, __FILE__, __LINE__);                   \
           goto error;                                           \
       }                                                         \
@@ -94,6 +94,8 @@ void pl_gpu_print_info(const struct pl_gpu *gpu, enum pl_log_level lev)
         }
 
         PL_MSG(gpu, lev, "      UUID: %s", buf);
+        PL_MSG(gpu, lev, "      PCI: %04x:%02x:%02x:%x",
+               gpu->pci.domain, gpu->pci.bus, gpu->pci.device, gpu->pci.function);
         PL_MSG(gpu, lev, "      buf export caps: 0x%x",
                (unsigned int) gpu->export_caps.buf);
         PL_MSG(gpu, lev, "      buf import caps: 0x%x",
@@ -176,10 +178,15 @@ void pl_gpu_verify_formats(struct pl_gpu *gpu)
             pl_assert(fmt->glsl_type);
             pl_assert(!fmt->opaque);
         }
+        pl_assert(!fmt->opaque || !(fmt->caps & PL_FMT_CAP_HOST_READABLE));
         if (fmt->internal_size != fmt->texel_size && !fmt->opaque)
             pl_assert(fmt->emulated);
         if (fmt->caps & (PL_FMT_CAP_STORABLE | PL_FMT_CAP_TEXEL_STORAGE))
             pl_assert(fmt->glsl_format);
+
+        // Assert uniqueness of name
+        for (int o = n + 1; o < gpu->num_formats; o++)
+            pl_assert(strcmp(fmt->name, gpu->formats[o]->name) != 0);
     }
 }
 
@@ -412,6 +419,8 @@ const struct pl_tex *pl_tex_create(const struct pl_gpu *gpu,
         require(params->w > 0);
         require(params->w <= gpu->limits.max_tex_1d_dim);
         require(!params->renderable);
+        require(!params->blit_src || gpu->caps & PL_GPU_CAP_BLITTABLE_1D_3D);
+        require(!params->blit_dst || gpu->caps & PL_GPU_CAP_BLITTABLE_1D_3D);
         break;
     case 2:
         require(params->w > 0 && params->h > 0);
@@ -424,11 +433,14 @@ const struct pl_tex *pl_tex_create(const struct pl_gpu *gpu,
         require(params->h <= gpu->limits.max_tex_3d_dim);
         require(params->d <= gpu->limits.max_tex_3d_dim);
         require(!params->renderable);
+        require(!params->blit_src || gpu->caps & PL_GPU_CAP_BLITTABLE_1D_3D);
+        require(!params->blit_dst || gpu->caps & PL_GPU_CAP_BLITTABLE_1D_3D);
         break;
     }
 
     const struct pl_fmt *fmt = params->format;
     require(fmt);
+    require(!params->host_readable || fmt->caps & PL_FMT_CAP_HOST_READABLE);
     require(!params->host_readable || !fmt->opaque);
     require(!params->host_writable || !fmt->opaque);
     require(!params->sampleable || fmt->caps & PL_FMT_CAP_SAMPLEABLE);
@@ -884,15 +896,15 @@ const char *pl_var_glsl_type_name(struct pl_var var)
         };                                              \
     }
 
-PL_VAR(SINT,  int,   1, 1);
-PL_VAR(UINT,  uint,  1, 1);
-PL_VAR(FLOAT, float, 1, 1);
-PL_VAR(FLOAT, vec2,  1, 2);
-PL_VAR(FLOAT, vec3,  1, 3);
-PL_VAR(FLOAT, vec4,  1, 4);
-PL_VAR(FLOAT, mat2,  2, 2);
-PL_VAR(FLOAT, mat3,  3, 3);
-PL_VAR(FLOAT, mat4,  4, 4);
+PL_VAR(SINT,  int,   1, 1)
+PL_VAR(UINT,  uint,  1, 1)
+PL_VAR(FLOAT, float, 1, 1)
+PL_VAR(FLOAT, vec2,  1, 2)
+PL_VAR(FLOAT, vec3,  1, 3)
+PL_VAR(FLOAT, vec4,  1, 4)
+PL_VAR(FLOAT, mat2,  2, 2)
+PL_VAR(FLOAT, mat3,  3, 3)
+PL_VAR(FLOAT, mat4,  4, 4)
 
 #undef PL_VAR
 
@@ -1187,7 +1199,7 @@ void pl_pass_run(const struct pl_gpu *gpu, const struct pl_pass_run_params *para
         pl_tex_invalidate(gpu, params->target);
 
     const struct pl_gpu_fns *impl = TA_PRIV(gpu);
-    return impl->pass_run(gpu, &new);
+    impl->pass_run(gpu, &new);
 
 error:
     return;
