@@ -19,6 +19,7 @@
 #define LIBPLACEBO_COLORSPACE_H_
 
 #include <stdbool.h>
+#include <stdint.h>
 #include <libplacebo/common.h>
 
 // The underlying color representation (e.g. RGB, XYZ or YCbCr)
@@ -145,6 +146,11 @@ void pl_color_repr_merge(struct pl_color_repr *orig,
 // depth will both be inferred as 8 bits for the purposes of this conversion.
 float pl_color_repr_normalize(struct pl_color_repr *repr);
 
+// Guesses the best color levels based on the specified color levels and
+// falling back to using the color system instead. YCbCr-like systems are
+// assumed to be TV range, otherwise this defaults to PC range.
+enum pl_color_levels pl_color_levels_guess(const struct pl_color_repr *repr);
+
 // The colorspace's primaries (gamut)
 enum pl_color_primaries {
     PL_COLOR_PRIM_UNKNOWN = 0,
@@ -246,14 +252,31 @@ struct pl_color_space {
     enum pl_color_light light;
 
     // The highest value that occurs in the signal, relative to the reference
-    // white. (0 = unknown)
+    // white, alternatively the brightest color value supported by a given
+    // color space. (0 = unknown)
     float sig_peak;
 
     // The average light level that occurs in the signal, relative to the
-    // reference white. (0 = unknown)
+    // reference white, alternatively the desired average brightness level of a
+    // given color space. (0 = unknown)
     float sig_avg;
+
+    // Additional scale factor for the signal's reference white. If this is set
+    // to a value higher than 1.0, then it's assumed that the signal's encoded
+    // reference white is assumed to be brighter than normal by this factor.
+    // This can be used to over- or under-expose content, especially HDR
+    // content. (0 = unknown)
+    //
+    // An example of where this could come in use is for using an SDR transfer
+    // function (e.g. PL_COLOR_TRC_LINEAR) to encode a HDR image or display.
+    float sig_scale;
 };
 
+// Returns whether or not a color space is considered as effectively HDR.
+// This is true when the effective signal peak is greater than the SDR
+// reference white (1.0), after application of the `sig_scale`.
+bool pl_color_space_is_hdr(struct pl_color_space csp);
+
 // Replaces unknown values in the first struct by those of the second struct.
 void pl_color_space_merge(struct pl_color_space *orig,
                           const struct pl_color_space *new);
