@@ -30,12 +30,9 @@
 // since it tries targetting a very small common subset of features that is
 // needed to implement libplacebo's rendering.
 //
-// NOTE: When speaking of "valid usage" or "must", invalid usage is assumed to
-// result in undefined behavior. (Typically, an error message is printed to
-// stderr and libplacebo aborts - but this is not guaranteed). So ensuring
-// valid API usage by the API user is absolutely crucial. If you want to be
-// freed from this reponsibility, use the higher level abstractions provided by
-// libplacebo alongside gpu.h.
+// NOTE: Most, but not all, parameter conditions (phrases such as "must" or
+// "valid usage" are explicitly tested and result in error messages followed by
+// graceful failure. Exceptions are noted where they exist.
 
 // Structure which wraps metadata describing GLSL capabilities.
 struct pl_glsl_desc {
@@ -49,6 +46,8 @@ enum {
     PL_GPU_CAP_COMPUTE          = 1 << 0, // supports compute shaders
     PL_GPU_CAP_PARALLEL_COMPUTE = 1 << 1, // supports multiple compute queues
     PL_GPU_CAP_INPUT_VARIABLES  = 1 << 2, // supports shader input variables
+    PL_GPU_CAP_MAPPED_BUFFERS   = 1 << 3, // supports host-mapped buffers
+    PL_GPU_CAP_BLITTABLE_1D_3D  = 1 << 4, // supports blittable 1D/3D textures
 };
 
 // Some `pl_gpu` operations allow sharing GPU resources with external APIs -
@@ -60,6 +59,13 @@ enum pl_handle_type {
     PL_HANDLE_FD        = (1 << 0), // `int fd` for POSIX-style APIs
     PL_HANDLE_WIN32     = (1 << 1), // `HANDLE` for win32 API
     PL_HANDLE_WIN32_KMT = (1 << 2), // `HANDLE` for pre-Windows-8 win32 API
+    PL_HANDLE_DMA_BUF   = (1 << 3), // 'int fd' for a dma_buf fd
+};
+
+struct pl_gpu_handle_caps {
+    pl_handle_caps tex;  // supported handles for `pl_tex` + `pl_shared_mem`
+    pl_handle_caps buf;  // supported handles for `pl_buf` + `pl_shared_mem`
+    pl_handle_caps sync; // supported handles for `pl_sync`
 };
 
 // Wrapper for the handle used to communicate a shared resource externally.
@@ -70,7 +76,7 @@ enum pl_handle_type {
 // types may be managed in different ways. (eg: CUDA takes ownership of an fd,
 // but does not take ownership of a win32 handle).
 union pl_handle {
-    int fd;         // PL_HANDLE_FD
+    int fd;         // PL_HANDLE_FD / PL_HANDLE_DMA_BUF
     void *handle;   // PL_HANDLE_WIN32 / PL_HANDLE_WIN32_KMT
 };
 
@@ -114,12 +120,18 @@ struct pl_gpu_limits {
     size_t align_tex_xfer_offset;   // optimal `pl_tex_transfer_params.buf_offset`
 };
 
+// Structure grouping PCI bus address fields for GPU devices
+struct pl_gpu_pci_address {
+    uint32_t domain;
+    uint32_t bus;
+    uint32_t device;
+    uint32_t function;
+};
+
 // Abstract device context which wraps an underlying graphics context and can
 // be used to dispatch rendering commands.
 struct pl_gpu {
     struct pl_context *ctx;  // the `pl_context` this GPU was initialized from
-    struct pl_gpu_fns *impl; // the underlying implementation (unique per GPU)
-    void *priv;
 
     pl_gpu_caps caps;            // PL_GPU_CAP_* bit field
     struct pl_glsl_desc glsl;    // GLSL version supported by this GPU
@@ -129,16 +141,18 @@ struct pl_gpu {
 
     // Fields relevant to external API interop. If the underlying device does
     // not support interop with other APIs, these will all be {0}.
-    uint8_t uuid[16];               // underlying device UUID
-    struct {
-        pl_handle_caps shared_mem;  // supported handles for `pl_shared_mem`
-        pl_handle_caps sync;        // supported handles for `pl_sync`
-    } handle_caps;
+    struct pl_gpu_handle_caps export_caps; // supported handles for exporting
+    struct pl_gpu_handle_caps import_caps; // supported handles for importing
+    uint8_t uuid[16];                      // underlying device UUID
 
     // Supported texture formats, in preference order. (If there are multiple
     // similar formats, the "better" ones come first)
     const struct pl_fmt **formats;
     int num_formats;
+
+    // PCI Bus address of the underlying device, to help with interop.
+    // This will only be filled in if interop is supported.
+    struct pl_gpu_pci_address pci;
 };
 
 // Helper function to align the given dimension (e.g. width or height) to a
@@ -165,22 +179,25 @@ enum pl_fmt_caps {
     PL_FMT_CAP_VERTEX       = 1 << 6, // may be used as a vertex attribute
     PL_FMT_CAP_TEXEL_UNIFORM = 1 << 7, // may be used as a texel uniform buffer
     PL_FMT_CAP_TEXEL_STORAGE = 1 << 8, // may be used as a texel storage buffer
+    PL_FMT_CAP_HOST_READABLE = 1 << 9, // may be used with `host_readable` textures
 
     // Notes:
     // - PL_FMT_CAP_LINEAR also implies PL_FMT_CAP_SAMPLEABLE
     // - PL_FMT_CAP_STORABLE also implies PL_GPU_CAP_COMPUTE
+    // - PL_FMT_CAP_BLENDABLE implies PL_FMT_CAP_RENDERABLE
     // - PL_FMT_CAP_VERTEX implies that the format is non-opaque
+    // - PL_FMT_CAP_HOST_READABLE implies that the format is non-opaque
 };
 
 // Structure describing a texel/vertex format.
 struct pl_fmt {
     const char *name;       // symbolic name for this format (e.g. rgba32f)
-    const void *priv;
 
     enum pl_fmt_type type;  // the format's data type and interpretation
     enum pl_fmt_caps caps;  // the features supported by this format
     int num_components;     // number of components for this format
     int component_depth[4]; // meaningful bits per component, texture precision
+    size_t internal_size;   // internal texel size (for blit compatibility)
 
     // This controls the relationship between the data as seen by the host and
     // the way it's interpreted by the texture. The host representation is
@@ -259,20 +276,38 @@ struct pl_tex_params {
     bool host_writable; // may be updated with pl_tex_upload()
     bool host_readable; // may be fetched with pl_tex_download()
 
+    // Note: For `blit_src`, `blit_dst`, the texture must either be
+    // 2-dimensional or `PL_GPU_CAP_BLITTABLE_1D_3D` must set.
+
     // The following capabilities are only relevant for textures which have
     // either sampleable or blit_src enabled.
     enum pl_tex_sample_mode sample_mode;
     enum pl_tex_address_mode address_mode;
 
+    // At most one of `export_handle` and `import_handle` can be set for a
+    // texture.
+
     // Setting this indicates that the memory backing this texture should be
     // shared with external APIs, If so, this must be exactly *one* of
-    // `pl_gpu.handle_caps.shared_mem`.
-    enum pl_handle_type handle_type;
+    // `pl_gpu.export_caps.tex`.
+    enum pl_handle_type export_handle;
 
-    // If non-NULL, the texture will be created with these contents. Using
-    // this does *not* require setting host_writable. Otherwise, the initial
-    // data is undefined.
+    // Setting this indicates that the memory backing this texture will be
+    // imported from an external API. If so, this must be exactly *one* of
+    // `pl_gpu.import_caps.tex`.
+    enum pl_handle_type import_handle;
+
+    // If the shared memory is being imported, the import handle must be
+    // specified here. Otherwise, this is ignored.
+    struct pl_shared_mem shared_mem;
+
+    // If non-NULL, the texture will be created with these contents (tightly
+    // packed). Using this does *not* require setting host_writable. Otherwise,
+    // the initial data is undefined.
     const void *initial_data;
+
+    // Arbitrary user data. libplacebo does not use this at all.
+    void *user_data;
 };
 
 static inline int pl_tex_params_dimension(const struct pl_tex_params params)
@@ -294,9 +329,8 @@ static inline int pl_tex_params_dimension(const struct pl_tex_params params)
 // and/or rendered to.
 struct pl_tex {
     struct pl_tex_params params;
-    void *priv;
 
-    // If `params.handle_type` is set, this structure references the shared
+    // If `params.export_handle` is set, this structure references the shared
     // memory backing this buffer, via the requested handle type.
     //
     // While this texture is not in an "exported" state, the contents of the
@@ -320,6 +354,10 @@ void pl_tex_destroy(const struct pl_gpu *gpu, const struct pl_tex **tex);
 // Even if the texture is not recreated, calling this function will still
 // invalidate the contents of the texture. (Note: Because of this,
 // `initial_data` may not be used with `pl_tex_recreate`. Doing so is an error)
+//
+// Note: If the `user_data` alone changes, this does not trigger a texture
+// recreation. In theory, this can be used to detect when the texture ended
+// up being recreated.
 bool pl_tex_recreate(const struct pl_gpu *gpu, const struct pl_tex **tex,
                      const struct pl_tex_params *params);
 
@@ -328,7 +366,7 @@ bool pl_tex_recreate(const struct pl_gpu *gpu, const struct pl_tex **tex,
 void pl_tex_invalidate(const struct pl_gpu *gpu, const struct pl_tex *tex);
 
 // Clear the dst texture with the given color (rgba). This is functionally
-// identical to a blit operation, which means dst->params.blit_dst must be
+// identical to a blit operation, which means `dst->params.blit_dst` must be
 // set.
 void pl_tex_clear(const struct pl_gpu *gpu, const struct pl_tex *dst,
                   const float color[4]);
@@ -336,14 +374,14 @@ void pl_tex_clear(const struct pl_gpu *gpu, const struct pl_tex *dst,
 // Copy a sub-rectangle from one texture to another. The source/dest regions
 // must be within the texture bounds. Areas outside the dest region are
 // preserved. The formats of the textures must be loosely compatible - which
-// essentially means that they must have the same texel size. Additionally,
+// essentially means that they must have the same `internal_size`. Additionally,
 // UINT textures can only be blitted to other UINT textures, and SINT textures
-// can only be blitted to other SINT textures. Finally, src.blit_src and
-// dst.blit_dst must be set, respectively.
+// can only be blitted to other SINT textures. Finally, `src.blit_src` and
+// `dst.blit_dst` must be set, respectively.
 //
 // The rectangles may be "flipped", which leads to the image being flipped
 // while blitting. If the src and dst rects have different sizes, the source
-// image will be scaled according to src->params.sample_mode. That said, the
+// image will be scaled according to `src->params.sample_mode`. That said, the
 // src and dst rects must be fully contained within the src/dst dimensions.
 void pl_tex_blit(const struct pl_gpu *gpu,
                  const struct pl_tex *dst, const struct pl_tex *src,
@@ -427,9 +465,10 @@ enum pl_buf_mem_type {
 struct pl_buf_params {
     enum pl_buf_type type;
     size_t size;        // size in bytes
-    bool host_mapped;   // create a persistent, RW mapping (pl_buf.data)
     bool host_writable; // contents may be updated via pl_buf_write()
     bool host_readable; // contents may be read back via pl_buf_read()
+    bool host_mapped;   // create a persistent, RW mapping (pl_buf.data)
+                        // (requires PL_GPU_CAP_MAPPED_BUFFERS)
 
     // Provide a hint for the memory type you want to use when allocating
     // this buffer's memory. Currently, this field is ignored for all buffer
@@ -444,13 +483,16 @@ struct pl_buf_params {
 
     // Setting this indicates that the memory backing this buffer should be
     // shared with external APIs, If so, this must be exactly *one* of
-    // `pl_gpu.handle_caps.shared_mem`.
+    // `pl_gpu.export_caps.buf`.
     enum pl_handle_type handle_type;
 
     // If non-NULL, the buffer will be created with these contents. Otherwise,
     // the initial data is undefined. Using this does *not* require setting
     // host_writable.
     const void *initial_data;
+
+    // Arbitrary user data. libplacebo does not use this at all.
+    void *user_data;
 };
 
 // A generic buffer, which can be used for multiple purposes (texture transfer,
@@ -465,7 +507,6 @@ struct pl_buf_params {
 struct pl_buf {
     struct pl_buf_params params;
     uint8_t *data; // for persistently mapped buffers, points to the first byte
-    void *priv;
 
     // If `params.handle_type` is set, this structure references the shared
     // memory backing this buffer, via the requested handle type.
@@ -498,6 +539,10 @@ const struct pl_buf *pl_buf_create(const struct pl_gpu *gpu,
 // external API's handle. Conversely, it *is* allowed on a buffer with
 // `params->host_mapped`, and the corresponding `buf->data` pointer *may*
 // change as a result of doing so.
+//
+// Note: If the `user_data` alone changes, this does not trigger a buffer
+// recreation. In theory, this can be used to detect when the texture ended
+// up being recreated.
 bool pl_buf_recreate(const struct pl_gpu *gpu, const struct pl_buf **buf,
                      const struct pl_buf_params *params);
 
@@ -505,12 +550,20 @@ void pl_buf_destroy(const struct pl_gpu *gpu, const struct pl_buf **buf);
 
 // Update the contents of a buffer, starting at a given offset (must be a
 // multiple of 4) and up to a given size, with the contents of *data.
+//
+// Note: This operation may write directly to the buffer, i.e. it is not
+// synchronized. Writing to an in-use buffer is undefined behavior. Use
+// `pl_buff_poll` to ensure write consistency.
 void pl_buf_write(const struct pl_gpu *gpu, const struct pl_buf *buf,
                   size_t buf_offset, const void *data, size_t size);
 
 // Read back the contents of a buffer, starting at a given offset (must be a
 // multiple of 4) and up to a given size, storing the data into *dest.
 // Returns whether successful.
+//
+// Note: This operation will never block, so reading from a buffer that is
+// currently being written to results in the read memory regions containing
+// undefined values. Use `pl_buf_poll` to ensure read consistency.
 bool pl_buf_read(const struct pl_gpu *gpu, const struct pl_buf *buf,
                  size_t buf_offset, void *dest, size_t size);
 
@@ -522,7 +575,8 @@ bool pl_buf_read(const struct pl_gpu *gpu, const struct pl_buf *buf,
 // operation that touches the buffer (e.g. pl_tex_upload, but also pl_buf_write
 // and pl_buf_read) will implicitly import the buffer back to libplacebo. Users
 // must ensure that all pending operations made by the external API are fully
-// completed before using it in libplacebo again.
+// completed before using it in libplacebo again. (Otherwise, the bahviour
+// is undefined)
 //
 // Please note that this function returning does not mean the memory is
 // immediately available as such. In general, it will mark a buffer as "in use"
@@ -562,10 +616,16 @@ bool pl_buf_export(const struct pl_gpu *gpu, const struct pl_buf *buf);
 // before returning. If set to 0, this function will never block, and only
 // returns the current status of the buffer. The actual precision of the
 // timeout may be significantly longer than one nanosecond, and has no upper
-// bound. This function does not provide hard latency guarantees.
+// bound. This function does not provide hard latency guarantees. This function
+// may also return at any time, even if the buffer is still in use. If the user
+// wishes to block until the buffer is definitely no longer in use, the
+// recommended usage is:
 //
-// Note: Performing multiple libplacebo operations at the same time is always
-// valid, for example it's perfectly valid to submit a `pl_tex_upload`
+// while (pl_buf_poll(gpu, buf, UINT64_MAX))
+//      ; // do nothing
+//
+// Note: Performing multiple libplacebo-internal operations at the same time is
+// always valid, for example it's perfectly valid to submit a `pl_tex_upload`
 // immediately followed by a `pl_tex_download` to the same buffer. However,
 // when this is the case, it's undefined as to when exactly the upload is
 // happening versus when exactly the download is happening. So in this example,
@@ -605,6 +665,7 @@ struct pl_var {
 const char *pl_var_glsl_type_name(struct pl_var var);
 
 // Helper functions for constructing the most common pl_vars.
+struct pl_var pl_var_int(const char *name);
 struct pl_var pl_var_uint(const char *name);
 struct pl_var pl_var_float(const char *name);
 struct pl_var pl_var_vec2(const char *name);
@@ -829,17 +890,12 @@ struct pl_pass_params {
     // The vertex shader itself.
     const char *vertex_shader;
 
-    // The target dummy texture this renderpass is intended to be used with.
-    // This doesn't have to be a real texture - the caller can also pass a
-    // blank pl_tex object, as long as target_dummy.params.format is set. The
-    // format must support PL_FMT_CAP_RENDERABLE, and if the target dummy is
-    // an actual texture, it must have `renderable` enabled.
-    //
-    // If you pass a real texture here, the GPU backend may be able to optimize
-    // the render pass better for the specific requirements of this texture.
-    // This does not change the semantics of pl_pass_run, just perhaps the
-    // performance. (The `priv` pointer will be cleared by pl_pass_create, so
-    // there is no risk of a dangling reference)
+    // The target texture this render pass is intended to be used with. This
+    // doesn't have to come from a real texture - the caller can also invent
+    // values or pass a blank struct, as long as `target_dummy.params.format`
+    // is set. The format must support `PL_FMT_CAP_RENDERABLE`. If any other
+    // fields are set, the GPU may be able to further optimize the render pass
+    // for this particular type of texture.
     struct pl_tex target_dummy;
 
     // Target blending mode. If this is NULL, blending is disabled. Otherwise,
@@ -860,7 +916,6 @@ struct pl_pass_params {
 // - the current values of all inputs
 struct pl_pass {
     struct pl_pass_params params;
-    void *priv;
 };
 
 // Compile a shader and create a render pass. This is a rare/expensive
@@ -934,7 +989,6 @@ void pl_pass_run(const struct pl_gpu *gpu, const struct pl_pass_run_params *para
 // semaphores - one to synchronize access in each direction.
 struct pl_sync {
     enum pl_handle_type handle_type;
-    void *priv;
 
     // This handle is signalled by the `pl_gpu`, and waited on by the user. It
     // fires when it is safe for the user to access the shared resource.
@@ -947,7 +1001,7 @@ struct pl_sync {
 
 // Create a synchronization object. Returns NULL on failure.
 //
-// `handle_type` must be exactly *one* of `pl_gpu.handle_caps.sync`, and
+// `handle_type` must be exactly *one* of `pl_gpu.export_caps.sync`, and
 // indicates which type of handle to generate for sharing this sync object.
 const struct pl_sync *pl_sync_create(const struct pl_gpu *gpu,
                                      enum pl_handle_type handle_type);
@@ -963,11 +1017,10 @@ void pl_sync_destroy(const struct pl_gpu *gpu,
                      const struct pl_sync **sync);
 
 // Initiates a texture export operation, allowing a texture to be accessed by
-// an external API. This is only valid for textures with `params.handle_type`.
-// Returns whether successful. After this operation successfully returns, it is
-// guaranteed that `sync->wait_handle` will eventually be signalled. For APIs
-// where this is relevant, the image layout should be specified as "general",
-// e.g. `GL_LAYOUT_GENERAL_EXT` for OpenGL.
+// an external API. Returns whether successful. After this operation
+// successfully returns, it is guaranteed that `sync->wait_handle` will
+// eventually be signalled. For APIs where this is relevant, the image layout
+// should be specified as "general", e.g. `GL_LAYOUT_GENERAL_EXT` for OpenGL.
 //
 // There is no corresponding "import" operation - the next operation that uses
 // a texture will implicitly import the texture. Valid API usage requires that
@@ -991,10 +1044,9 @@ bool pl_tex_export(const struct pl_gpu *gpu, const struct pl_tex *tex,
 // retroactively.
 //
 // The only time this might be beneficial to call explicitly is if you're doing
-// lots of offline processing where you submit a bunch of work and then
-// use asynchronous texture downloads (via pl_tex_download) to retrieve the
-// results. In that case you should call this function after each work item
-// to ensure good parallelism between them.
+// lots of offline processing, i.e. you aren't rendering to a swapchain but to
+// textures that you download from again. In that case you should call this
+// function after each "work item" to ensure good parallelism between them.
 //
 // It's worth noting that this function may block if you're over-feeding the
 // GPU without waiting for existing results to finish.
