@@ -21,6 +21,7 @@
 #include <libplacebo/colorspace.h>
 #include <libplacebo/filters.h>
 #include <libplacebo/gpu.h>
+#include <libplacebo/shaders/av1.h>
 #include <libplacebo/shaders/colorspace.h>
 #include <libplacebo/shaders/sampling.h>
 #include <libplacebo/swapchain.h>
@@ -81,6 +82,9 @@ struct pl_render_params {
 
     // Configures the settings used to deband source textures. Leaving this as
     // NULL disables debanding.
+    //
+    // Note: The `deband_params.grain` setting is automatically adjusted to
+    // prevent blowing up on HDR sources. The user need not account for this.
     const struct pl_deband_params *deband_params;
 
     // Configures the settings used to sigmoidize the image before upscaling.
@@ -92,6 +96,11 @@ struct pl_render_params {
     // desaturation, etc. If NULL, defaults to &pl_color_adjustment_neutral.
     const struct pl_color_adjustment *color_adjustment;
 
+    // Configures the settings used to detect the peak of the source content,
+    // for HDR sources. Has no effect on SDR content. If NULL, peak detection
+    // is disabled.
+    const struct pl_peak_detect_params *peak_detect_params;
+
     // Configures the settings used to tone map from HDR to SDR, or from higher
     // gamut to standard gamut content. If NULL, defaults to
     // `&pl_color_map_default_params`.
@@ -132,6 +141,12 @@ struct pl_render_params {
     // already disabled if the overlay texture does not need to be scaled.
     bool disable_overlay_sampling;
 
+    // Allows the peak detection result to be delayed by up to a single frame,
+    // which can sometimes (not always) allow skipping some otherwise redundant
+    // sampling work. Only relevant when peak detection is active (i.e.
+    // params->peak_detect_params is set and the source is HDR).
+    bool allow_delayed_peak_detect;
+
     // --- Performance tuning / debugging options
     // These may affect performance or may make debugging problems easier,
     // but shouldn't have any effect on the quality.
@@ -169,6 +184,16 @@ struct pl_render_params {
     // unnecessary. This is slower, but may improve the quality of the gamut
     // reduction step, if one is performed.
     bool force_3dlut;
+
+    // Forces the use of dithering, even when rendering to 16-bit FBOs. This is
+    // generally pretty pointless because most 16-bit FBOs have high enough
+    // depth that rounding errors are below the human perception threshold,
+    // but this can be used to test the dither code.
+    bool force_dither;
+
+    // Completely overrides the use of FBOs, as if there were no renderable
+    // texture format available. This disables most features.
+    bool disable_fbos;
 };
 
 // This contains the default/recommended options for reasonable image quality,
@@ -333,6 +358,10 @@ struct pl_image {
     // implementations. So in this example, the 18x12 chroma plane would get
     // treated by libplacebo as an oversized chroma plane - i.e. the plane
     // would get sampled as if it was 17.5 pixels wide and 11.5 pixels large.
+
+    // Associated AV1 grain params (see <libplacebo/shaders/av1.h>). This is
+    // entirely optional, the default of {0} corresponds to no extra grain.
+    struct pl_av1_grain_data av1_grain;
 };
 
 // Represents the target of a rendering operation
