@@ -75,13 +75,16 @@ struct pl_renderer {
     bool disable_overlay;       // disable rendering overlays
     bool disable_3dlut;         // disable usage of a 3DLUT
     bool disable_peak_detect;   // disable peak detection shader
+    bool disable_grain;         // disable AV1 grain code
 
     // Shader resource objects and intermediate textures (FBOs)
     struct pl_shader_obj *peak_detect_state;
     struct pl_shader_obj *dither_state;
     struct pl_shader_obj *lut3d_state;
+    struct pl_shader_obj *grain_state;
     const struct pl_tex *main_scale_fbo;
     const struct pl_tex *deband_fbos[PLANE_COUNT];
+    const struct pl_tex *grain_fbos[PLANE_COUNT];
     const struct pl_tex *output_fbo;
     struct sampler samplers[SCALER_COUNT];
     struct sampler *osd_samplers;
@@ -176,12 +179,15 @@ void pl_renderer_destroy(struct pl_renderer **p_rr)
     pl_tex_destroy(rr->gpu, &rr->main_scale_fbo);
     for (int i = 0; i < PL_ARRAY_SIZE(rr->deband_fbos); i++)
         pl_tex_destroy(rr->gpu, &rr->deband_fbos[i]);
+    for (int i = 0; i < PL_ARRAY_SIZE(rr->grain_fbos); i++)
+        pl_tex_destroy(rr->gpu, &rr->grain_fbos[i]);
     pl_tex_destroy(rr->gpu, &rr->output_fbo);
 
     // Free all shader resource objects
     pl_shader_obj_destroy(&rr->peak_detect_state);
     pl_shader_obj_destroy(&rr->dither_state);
     pl_shader_obj_destroy(&rr->lut3d_state);
+    pl_shader_obj_destroy(&rr->grain_state);
 
     // Free all samplers
     for (int i = 0; i < PL_ARRAY_SIZE(rr->samplers); i++)
@@ -210,6 +216,8 @@ const struct pl_render_params pl_render_default_params = {
     .dither_params      = &pl_dither_default_params,
 };
 
+#define FBOFMT (params->disable_fbos ? NULL : rr->fbofmt)
+
 // Represents a "in-flight" image, which is a shader that's in the process of
 // producing some sort of image
 struct img {
@@ -313,7 +321,7 @@ static struct sampler_info sample_src_info(struct pl_renderer *rr,
     }
 
     bool is_linear = src->tex->params.sample_mode == PL_TEX_SAMPLE_LINEAR;
-    if (!rr->fbofmt || rr->disable_sampling || !info.config) {
+    if (!FBOFMT || rr->disable_sampling || !info.config) {
         info.type = SAMPLER_DIRECT;
     } else {
         info.type = SAMPLER_COMPLEX;
@@ -397,7 +405,7 @@ static void dispatch_sampler(struct pl_renderer *rr, struct pl_shader *sh,
         };
 
         struct pl_sample_src src2 = *src;
-        src2.tex = finalize_img(rr, &img, rr->fbofmt, sep_fbo);
+        src2.tex = finalize_img(rr, &img, FBOFMT, sep_fbo);
         ok = src2.tex && pl_shader_sample_ortho(sh, PL_SEP_HORIZ, &src2, &fparams);
     }
 
@@ -529,6 +537,7 @@ enum {
 
 static int deband_src(struct pl_renderer *rr, struct pl_shader *psh,
                       struct pl_sample_src *psrc, const struct pl_tex **fbo,
+                      const struct pl_image *image,
                       const struct pl_render_params *params)
 {
     if (rr->disable_debanding || !params->deband_params)
@@ -572,7 +581,16 @@ static int deband_src(struct pl_renderer *rr, struct pl_shader *psh,
         }
     }
 
-    pl_shader_deband(sh, src, params->deband_params);
+    // Divide the deband grain scale by the effective current colorspace nominal
+    // peak, to make sure the output intensity of the grain is as independent
+    // of the source as possible, even though it happens this early in the
+    // process (well before any linearization / output adaptation)
+    struct pl_deband_params dparams = *params->deband_params;
+    float scale = pl_color_transfer_nominal_peak(image->color.transfer)
+                * image->color.sig_scale;
+    dparams.grain /= scale;
+
+    pl_shader_deband(sh, src, &dparams);
 
     if (deband_scales)
         return DEBAND_SCALED;
@@ -583,7 +601,7 @@ static int deband_src(struct pl_renderer *rr, struct pl_shader *psh,
         .h  = src->new_h,
     };
 
-    const struct pl_tex *new = finalize_img(rr, &img, rr->fbofmt, fbo);
+    const struct pl_tex *new = finalize_img(rr, &img, FBOFMT, fbo);
     if (!new) {
         PL_ERR(rr, "Failed dispatching debanding shader.. disabling debanding!");
         rr->disable_debanding = true;
@@ -610,7 +628,7 @@ static void hdr_update_peak(struct pl_renderer *rr, struct pl_shader *sh,
     if (rr->disable_compute || rr->disable_peak_detect)
         goto cleanup;
 
-    if (!rr->fbofmt && !params->allow_delayed_peak_detect) {
+    if (!FBOFMT && !params->allow_delayed_peak_detect) {
         PL_WARN(rr, "Disabling peak detection because "
                 "`allow_delayed_peak_detect` is false, but lack of FBOs "
                 "forces the result to be delayed.");
@@ -703,8 +721,8 @@ static bool pass_read_image(struct pl_renderer *rr, struct pass_state *pass,
     float scale = pl_color_repr_normalize(&repr);
 
     for (int i = 0; i < image->num_planes; i++) {
-        struct pl_shader *psh = pl_dispatch_begin_ex(rr->dp, true);
         struct pl_plane *plane = &planes[i];
+        float plane_scale = scale;
 
         // Compute the source shift/scale relative to the reference size
         float pw = plane->texture->params.w,
@@ -720,10 +738,76 @@ static bool pass_read_image(struct pl_renderer *rr, struct pass_state *pass,
         float rrx = rx >= 1 ? roundf(rx) : 1.0 / roundf(1.0 / rx),
               rry = ry >= 1 ? roundf(ry) : 1.0 / roundf(1.0 / ry);
 
+        // Perform AV1 grain first (before debanding) for several reasons:
+        // 1. It's closer to the intent of the spec to add grain first
+        // 2. Debanding implicitly crops, which screws with the AV1 grain code
+        // 3. Debanding may finalize the image (work-aroundable but annoying)
+
+        struct pl_av1_grain_params grain_params = {
+            .data = image->av1_grain,
+            .luma_tex = refplane->texture,
+            .repr = repr,
+            .channels = {PL_CHANNEL_NONE, PL_CHANNEL_NONE, PL_CHANNEL_NONE},
+            .sub_x = ilogbf(rrx),
+            .sub_y = ilogbf(rry),
+        };
+
+        for (int c = 0; c < plane->components; c++) {
+            int idx = plane->texture->params.format->sample_order[c];
+            if (idx < 0 || idx >= PL_ARRAY_SIZE(grain_params.channels))
+                continue;
+            grain_params.channels[idx] = plane->component_mapping[c];
+        }
+
+        const struct pl_fmt *grain_fmt = plane->texture->params.format;
+        if (!(grain_fmt->caps & PL_FMT_CAP_RENDERABLE))
+            grain_fmt = FBOFMT;
+
+        bool needs_grain = !rr->disable_grain && pl_needs_av1_grain(&grain_params);
+        if (needs_grain && !grain_fmt) {
+            PL_ERR(rr, "AV1 grain required but no renderable format available.. "
+                  "disabling!");
+            rr->disable_grain = true;
+            needs_grain = false;
+        }
+
+        if (needs_grain) {
+            struct pl_shader *gsh = pl_dispatch_begin_ex(rr->dp, false);
+            const struct pl_tex *new_tex;
+            bool ok = pl_shader_sample_direct(gsh, &(struct pl_sample_src) {
+                .tex = plane->texture,
+                .scale = plane_scale,
+            });
+
+            if (ok)
+                ok = pl_shader_av1_grain(gsh, &rr->grain_state, &grain_params);
+
+            if (ok) {
+                struct img grain_img = {
+                    .sh = gsh,
+                    .w = plane->texture->params.w,
+                    .h = plane->texture->params.h,
+                };
+
+                new_tex = finalize_img(rr, &grain_img, grain_fmt, &rr->grain_fbos[i]);
+                ok = !!new_tex;
+                gsh = NULL;
+            }
+
+            if (ok) {
+                plane->texture = new_tex;
+                plane_scale = 1.0;
+            } else {
+                PL_ERR(rr, "Failed applying AV1 grain.. disabling!");
+                rr->disable_grain = true;
+                pl_dispatch_abort(rr->dp, &gsh);
+            }
+        }
+
         struct pl_sample_src src = {
             .tex        = plane->texture,
             .components = plane->components,
-            .scale      = scale,
+            .scale      = plane_scale,
             .new_w      = target_w,
             .new_h      = target_h,
             .rect       = {
@@ -734,7 +818,8 @@ static bool pass_read_image(struct pl_renderer *rr, struct pass_state *pass,
             },
         };
 
-        if (deband_src(rr, psh, &src, &rr->deband_fbos[i], params) != DEBAND_SCALED)
+        struct pl_shader *psh = pl_dispatch_begin_ex(rr->dp, true);
+        if (deband_src(rr, psh, &src, &rr->deband_fbos[i], image, params) != DEBAND_SCALED)
             dispatch_sampler(rr, psh, &rr->samplers[i], params, &src);
 
         ident_t sub = sh_subpass(sh, psh);
@@ -797,14 +882,14 @@ static bool pass_scale_main(struct pl_renderer *rr, struct pass_state *pass,
                             const struct pl_render_target *target,
                             const struct pl_render_params *params)
 {
-    if (!rr->fbofmt) {
+    if (!FBOFMT) {
         PL_TRACE(rr, "Skipping main scaler (no FBOs)");
         return true;
     }
 
     struct img *img = &pass->cur_img;
     struct pl_sample_src src = {
-        .tex        = &(struct pl_tex) { .params = img_params(rr, img, rr->fbofmt) },
+        .tex        = &(struct pl_tex) { .params = img_params(rr, img, FBOFMT) },
         .components = img->comps,
         .new_w      = abs(pl_rect_w(target->dst_rect)),
         .new_h      = abs(pl_rect_h(target->dst_rect)),
@@ -848,7 +933,7 @@ static bool pass_scale_main(struct pl_renderer *rr, struct pass_state *pass,
     if (use_sigmoid)
         pl_shader_sigmoidize(img->sh, params->sigmoid_params);
 
-    src.tex = finalize_img(rr, img, rr->fbofmt, &rr->main_scale_fbo);
+    src.tex = finalize_img(rr, img, FBOFMT, &rr->main_scale_fbo);
     if (!src.tex)
         return false;
 
@@ -951,7 +1036,7 @@ fallback:
 
     bool is_comp = pl_shader_is_compute(sh);
     if (is_comp && !fbo->params.storable) {
-        bool ok = finalize_img(rr, &pass->cur_img, rr->fbofmt, &rr->output_fbo);
+        bool ok = finalize_img(rr, &pass->cur_img, FBOFMT, &rr->output_fbo);
         if (!ok) {
             PL_ERR(rr, "Failed dispatching compute shader to intermediate FBO?");
             return false;
@@ -978,7 +1063,7 @@ fallback:
         int depth = PL_DEF(target->repr.bits.sample_depth, fmt_depth);
 
         // Ignore dithering for >16-bit FBOs, since it's pretty pointless
-        if (depth <= 16)
+        if (depth <= 16 || params->force_dither)
             pl_shader_dither(sh, depth, &rr->dither_state, params->dither_params);
     }
 
@@ -1048,7 +1133,7 @@ bool pl_render_image(struct pl_renderer *rr, const struct pl_image *pimage,
 
     // If we don't have FBOs available, simulate the on-image overlays at
     // this stage
-    if (image.num_overlays > 0 && !rr->fbofmt) {
+    if (image.num_overlays > 0 && !FBOFMT) {
         float rx = pl_rect_w(target.dst_rect) / pl_rect_w(image.src_rect),
               ry = pl_rect_h(target.dst_rect) / pl_rect_h(image.src_rect);
 
