@@ -1,18 +1,27 @@
 /*
- * This file is part of libplacebo.
+ * This file is part of libplacebo, which is normally licensed under the terms
+ * of the LGPL v2.1+. However, this file (av1.c) is also available under the
+ * terms of the more permissive MIT license:
  *
- * libplacebo is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
+ * Copyright (c) 2018-2019 Niklas Haas
  *
- * libplacebo is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
+ * Permission is hereby granted, free of charge, to any person obtaining a copy
+ * of this software and associated documentation files (the "Software"), to deal
+ * in the Software without restriction, including without limitation the rights
+ * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+ * copies of the Software, and to permit persons to whom the Software is
+ * furnished to do so, subject to the following conditions:
  *
- * You should have received a copy of the GNU Lesser General Public
- * License along with libplacebo. If not, see <http://www.gnu.org/licenses/>.
+ * The above copyright notice and this permission notice shall be included in all
+ * copies or substantial portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+ * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+ * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #include "shaders.h"
@@ -269,9 +278,10 @@ struct grain_scale {
     float grain_scale;
 };
 
-static struct grain_scale get_grain_scale(const struct pl_grain_params *params)
+static struct grain_scale get_grain_scale(const struct pl_av1_grain_params *params)
 {
-    int bit_depth = params->repr.bits.color_depth;
+    int bit_depth = PL_DEF(params->repr.bits.color_depth, 8);
+    pl_assert(bit_depth >= 8);
     struct grain_scale ret = {
         .grain_center = 128 << (bit_depth - 8),
     };
@@ -302,12 +312,14 @@ static struct grain_scale get_grain_scale(const struct pl_grain_params *params)
 // Generates the basic grain table (LumaGrain in the spec).
 static void generate_grain_y(float out[GRAIN_HEIGHT_LUT][GRAIN_WIDTH_LUT],
                              int16_t buf[GRAIN_HEIGHT][GRAIN_WIDTH],
-                             const struct pl_grain_params *params)
+                             const struct pl_av1_grain_params *params)
 {
+    const struct pl_av1_grain_data *data = &params->data;
     struct grain_scale scale = get_grain_scale(params);
-    uint16_t seed = params->grain_seed;
-    int bit_depth = params->repr.bits.color_depth;
-    int shift = 12 - bit_depth + params->grain_scale_shift;
+    uint16_t seed = data->grain_seed;
+    int bit_depth = PL_DEF(params->repr.bits.color_depth, 8);
+    int shift = 12 - bit_depth + data->grain_scale_shift;
+    pl_assert(shift >= 0);
 
     for (int y = 0; y < GRAIN_HEIGHT; y++) {
         for (int x = 0; x < GRAIN_WIDTH; x++) {
@@ -317,11 +329,11 @@ static void generate_grain_y(float out[GRAIN_HEIGHT_LUT][GRAIN_WIDTH_LUT],
     }
 
     const int ar_pad = 3;
-    int ar_lag = params->ar_coeff_lag;
+    int ar_lag = data->ar_coeff_lag;
 
     for (int y = ar_pad; y < GRAIN_HEIGHT; y++) {
         for (int x = ar_pad; x < GRAIN_WIDTH - ar_pad; x++) {
-            const int8_t *coeff = params->ar_coeffs_y;
+            const int8_t *coeff = data->ar_coeffs_y;
             int sum = 0;
             for (int dy = -ar_lag; dy <= 0; dy++) {
                 for (int dx = -ar_lag; dx <= ar_lag; dx++) {
@@ -331,7 +343,7 @@ static void generate_grain_y(float out[GRAIN_HEIGHT_LUT][GRAIN_WIDTH_LUT],
                 }
             }
 
-            int16_t grain = buf[y][x] + round2(sum, params->ar_coeff_shift);
+            int16_t grain = buf[y][x] + round2(sum, data->ar_coeff_shift);
             grain = PL_MAX(scale.grain_min, PL_MIN(scale.grain_max, grain));
             buf[y][x] = grain;
         }
@@ -348,13 +360,15 @@ static void generate_grain_y(float out[GRAIN_HEIGHT_LUT][GRAIN_WIDTH_LUT],
 static void generate_grain_uv(float *out, int16_t buf[GRAIN_HEIGHT][GRAIN_WIDTH],
                               const int16_t buf_y[GRAIN_HEIGHT][GRAIN_WIDTH],
                               enum pl_channel channel,
-                              const struct pl_grain_params *params)
+                              const struct pl_av1_grain_params *params)
 {
+    const struct pl_av1_grain_data *data = &params->data;
     struct grain_scale scale = get_grain_scale(params);
-    int bit_depth = params->repr.bits.color_depth;
-    int shift = 12 - bit_depth + params->grain_scale_shift;
+    int bit_depth = PL_DEF(params->repr.bits.color_depth, 8);
+    int shift = 12 - bit_depth + data->grain_scale_shift;
+    pl_assert(shift >= 0);
 
-    uint16_t seed = params->grain_seed;
+    uint16_t seed = data->grain_seed;
     if (channel == PL_CHANNEL_CB) {
         seed ^= 0xb524;
     } else if (channel == PL_CHANNEL_CR) {
@@ -365,8 +379,8 @@ static void generate_grain_uv(float *out, int16_t buf[GRAIN_HEIGHT][GRAIN_WIDTH]
     int chromaH = params->sub_y ? SUB_GRAIN_HEIGHT : GRAIN_HEIGHT;
 
     const int8_t *coeffs[] = {
-        [PL_CHANNEL_CB] = params->ar_coeffs_uv[0],
-        [PL_CHANNEL_CR] = params->ar_coeffs_uv[1],
+        [PL_CHANNEL_CB] = data->ar_coeffs_uv[0],
+        [PL_CHANNEL_CR] = data->ar_coeffs_uv[1],
     };
 
     for (int y = 0; y < chromaH; y++) {
@@ -377,7 +391,7 @@ static void generate_grain_uv(float *out, int16_t buf[GRAIN_HEIGHT][GRAIN_WIDTH]
     }
 
     const int ar_pad = 3;
-    int ar_lag = params->ar_coeff_lag;
+    int ar_lag = data->ar_coeff_lag;
 
     for (int y = ar_pad; y < chromaH; y++) {
         for (int x = ar_pad; x < chromaW - ar_pad; x++) {
@@ -388,7 +402,7 @@ static void generate_grain_uv(float *out, int16_t buf[GRAIN_HEIGHT][GRAIN_WIDTH]
                     // For the final (current) pixel, we need to add in the
                     // contribution from the luma grain texture
                     if (!dx && !dy) {
-                        if (!params->num_points_y)
+                        if (!data->num_points_y)
                             break;
                         int luma = 0;
                         int lumaX = ((x - ar_pad) << params->sub_x) + ar_pad;
@@ -407,7 +421,7 @@ static void generate_grain_uv(float *out, int16_t buf[GRAIN_HEIGHT][GRAIN_WIDTH]
                 }
             }
 
-            int16_t grain = buf[y][x] + round2(sum, params->ar_coeff_shift);
+            int16_t grain = buf[y][x] + round2(sum, data->ar_coeff_shift);
             grain = PL_MAX(scale.grain_min, PL_MIN(scale.grain_max, grain));
             buf[y][x] = grain;
         }
@@ -427,10 +441,10 @@ static void generate_grain_uv(float *out, int16_t buf[GRAIN_HEIGHT][GRAIN_WIDTH]
 }
 
 static void generate_offsets(uint32_t *buf, int offsets_x, int offsets_y,
-                             const struct pl_grain_params *params)
+                             const struct pl_av1_grain_data *data)
 {
     for (int y = 0; y < offsets_y; y++) {
-        uint16_t state = params->grain_seed;
+        uint16_t state = data->grain_seed;
         state ^= ((y * 37 + 178) & 0xFF) << 8;
         state ^= ((y * 173 + 105) & 0xFF);
 
@@ -445,10 +459,10 @@ static void generate_offsets(uint32_t *buf, int offsets_x, int offsets_y,
             // Encode four offsets into a single 32-bit integer for
             // the convenience of the GPU. That way only one SSBO read is
             // required for the entire block.
-            *offsets = (val_tl << OFFSET_TL)
-                     | (val_t  << OFFSET_T)
-                     | (val_l  << OFFSET_L)
-                     | (val    << OFFSET_N);
+            *offsets = ((uint32_t) val_tl << OFFSET_TL)
+                     | ((uint32_t) val_t  << OFFSET_T)
+                     | ((uint32_t) val_l  << OFFSET_L)
+                     | ((uint32_t) val    << OFFSET_N);
         }
     }
 }
@@ -460,10 +474,10 @@ static void generate_scaling(void *priv, float *data, int w, int h, int d)
     struct {
         int num;
         uint8_t (*points)[2];
-        const struct pl_grain_params *params;
+        const struct pl_av1_grain_data *data;
     } *ctx = priv;
 
-    float range = 1 << ctx->params->scaling_shift;
+    float range = 1 << ctx->data->scaling_shift;
 
     // Fill up the preceding entries with the initial value
     for (int i = 0; i < ctx->points[0][0]; i++)
@@ -488,7 +502,7 @@ static void generate_scaling(void *priv, float *data, int w, int h, int d)
 }
 
 static void sample(struct pl_shader *sh, enum offset off, enum pl_channel c,
-                   const struct pl_grain_params *params)
+                   const struct pl_av1_grain_params *params)
 {
     static const struct { int dx, dy; } delta[] = {
         [OFFSET_TL] = {1, 1},
@@ -504,7 +518,7 @@ static void sample(struct pl_shader *sh, enum offset off, enum pl_channel c,
     GLSL("offset = %du * uvec2((data >> %d) & 0xFu,   \n"
          "                     (data >> %d) & 0xFu);  \n"
          "pos = offset + local_id.xy + uvec2(%d, %d); \n"
-         "val = grain_%s[ pos.y * %d + pos.x ];       \n",
+         "val = grain_%s[ pos.y * %du + pos.x ];      \n",
          luma ? 2 : 1, off + 4, off,
          (BLOCK_SIZE >> sub_x) * delta[off].dx,
          (BLOCK_SIZE >> sub_y) * delta[off].dy,
@@ -512,15 +526,16 @@ static void sample(struct pl_shader *sh, enum offset off, enum pl_channel c,
 }
 
 static void get_grain_for_channel(struct pl_shader *sh, enum pl_channel c,
-                                  const struct pl_grain_params *params)
+                                  const struct pl_av1_grain_params *params)
 {
     if (c == PL_CHANNEL_NONE)
         return;
 
+    const struct pl_av1_grain_data *data = &params->data;
     sample(sh, OFFSET_N, c, params);
     GLSL("grain = val; \n");
 
-    if (params->overlap) {
+    if (data->overlap) {
         int sub_x = c == PL_CHANNEL_Y ? 0 : params->sub_x;
         int sub_y = c == PL_CHANNEL_Y ? 0 : params->sub_y;
         const char *weights[] = { "vec2(27.0, 17.0)", "vec2(23.0, 22.0)" };
@@ -581,7 +596,8 @@ struct sh_grain_obj {
     int num_offsets;
     int chroma_lut_size;
     uint32_t *offsets;
-    struct pl_grain_params params;
+    struct pl_av1_grain_data data;
+    int sub_x, sub_y;
 
     // Space to store the temporary arrays, reused
     float grain[GRAIN_HEIGHT_LUT][GRAIN_WIDTH_LUT];
@@ -598,43 +614,67 @@ static void sh_grain_uninit(const struct pl_gpu *gpu, void *ptr)
     *obj = (struct sh_grain_obj) {0};
 }
 
-void pl_shader_av1_grain(struct pl_shader *sh,
-                         struct pl_shader_obj **grain_state,
-                         const enum pl_channel channels[3],
-                         const struct pl_tex *luma_tex,
-                         const struct pl_grain_params *params)
+bool pl_needs_av1_grain(const struct pl_av1_grain_params *params)
 {
-    bool has_luma = params->num_points_y > 0;
-    bool has_chroma = params->num_points_uv[0] > 0 ||
-                      params->num_points_uv[1] > 0 ||
-                      params->chroma_scaling_from_luma;
+    const struct pl_av1_grain_data *data = &params->data;
+    bool has_luma = data->num_points_y > 0;
+    bool has_chroma = data->num_points_uv[0] > 0 ||
+                      data->num_points_uv[1] > 0 ||
+                      data->chroma_scaling_from_luma;
 
     int chmask = 0;
     for (int i = 0; i < 3; i++) {
-        if (channels[i] != PL_CHANNEL_NONE)
-            chmask |= 1 << channels[i];
+        if (params->channels[i] != PL_CHANNEL_NONE)
+            chmask |= 1 << params->channels[i];
     }
 
     bool is_luma = chmask & 0x1;
     bool is_chroma = chmask & 0x6;
-    if (!(has_luma && is_luma) && !(has_chroma && is_chroma))
-        return; // nothing to do
+    return (has_luma && is_luma) || (has_chroma && is_chroma);
+}
 
-    if (!is_luma && !luma_tex) {
-        PL_ERR(sh, "When pl_shader_av1_grain is used on a chroma plane, "
-               "the corresponding luma texture must be given as `luma_tex`.");
-        return;
+bool pl_shader_av1_grain(struct pl_shader *sh,
+                         struct pl_shader_obj **grain_state,
+                         const struct pl_av1_grain_params *params)
+{
+    pl_assert(params->luma_tex);
+    int width = params->luma_tex->params.w,
+        height = params->luma_tex->params.h;
+
+    if (!pl_needs_av1_grain(params)) {
+        PL_DEBUG(sh, "pl_shader_av1_grain called but no AV1 grain needs to be "
+                 "applied, consider testing with `pl_needs_av1_grain` first!");
+        return true; // nothing to do
     }
 
+    const struct pl_av1_grain_data *data = &params->data;
+    bool has_luma = data->num_points_y > 0;
+    bool has_chroma = data->num_points_uv[0] > 0 ||
+                      data->num_points_uv[1] > 0 ||
+                      data->chroma_scaling_from_luma;
+
+    int chmask = 0;
+    for (int i = 0; i < 3; i++) {
+        if (params->channels[i] != PL_CHANNEL_NONE)
+            chmask |= 1 << params->channels[i];
+    }
+
+    bool is_luma = chmask & 0x1;
+    bool is_chroma = chmask & 0x6;
     if (is_luma && is_chroma && (params->sub_x || params->sub_y)) {
         PL_ERR(sh, "pl_shader_av1_grain can't be called on luma and chroma "
                "at the same time with subsampled chroma! Please only call "
                "this function on the correctly sized textures.");
-        return;
+        return false;
     }
 
     if (!sh_require(sh, PL_SHADER_SIG_COLOR, 0, 0))
-        return;
+        return false;
+
+    if (SH_GPU(sh)->glsl.version < 130) {
+        PL_ERR(sh, "pl_shader_av1_grain requires GLSL >= 130!");
+        return false;
+    }
 
     int bw = BLOCK_SIZE >> (is_chroma ? params->sub_x : 0);
     int bh = BLOCK_SIZE >> (is_chroma ? params->sub_y : 0);
@@ -644,10 +684,10 @@ void pl_shader_av1_grain(struct pl_shader *sh,
     obj = SH_OBJ(sh, grain_state, PL_SHADER_OBJ_AV1_GRAIN,
                  struct sh_grain_obj, sh_grain_uninit);
     if (!obj)
-        return;
+        return false;
 
-    int offsets_x = PL_ALIGN2(params->width,  128) / 32;
-    int offsets_y = PL_ALIGN2(params->height, 128) / 32;
+    int offsets_x = PL_ALIGN2(width,  128) / 32;
+    int offsets_y = PL_ALIGN2(height, 128) / 32;
 
     int chroma_lut_size = (GRAIN_WIDTH_LUT >> params->sub_x)
                         * (GRAIN_HEIGHT_LUT >> params->sub_y);
@@ -656,25 +696,27 @@ void pl_shader_av1_grain(struct pl_shader *sh,
     // luma or only related to chroma and skip updating parts of the SSBO,
     // but this is probably not worth it since the grain_seed is expected
     // to change per frame anyway.
-    bool needs_update = memcmp(params, &obj->params, sizeof(*params)) != 0;
+    bool needs_update = memcmp(data, &obj->data, sizeof(*data)) != 0 ||
+                        params->sub_x != obj->sub_x ||
+                        params->sub_y != obj->sub_y;
 
     // For the scaling LUTs, we assume they'll be relatively constant
     // throughout the video so doing some extra work to avoid reinitializing
     // them constantly is probably worth it. Probably.
     bool scaling_changed = false;
-    if (has_luma || params->chroma_scaling_from_luma) {
-        scaling_changed |= params->num_points_y != obj->params.num_points_y;
-        scaling_changed |= memcmp(params->points_y, obj->params.points_y,
-                                  sizeof(params->points_y));
+    if (has_luma || data->chroma_scaling_from_luma) {
+        scaling_changed |= data->num_points_y != obj->data.num_points_y;
+        scaling_changed |= memcmp(data->points_y, obj->data.points_y,
+                                  sizeof(data->points_y));
     }
 
-    if (has_chroma && !params->chroma_scaling_from_luma) {
+    if (has_chroma && !data->chroma_scaling_from_luma) {
         for (int i = 0; i < 2; i++) {
-            scaling_changed |= params->num_points_uv[i] !=
-                               obj->params.num_points_uv[i];
-            scaling_changed |= memcmp(params->points_uv[i],
-                                      obj->params.points_uv[i],
-                                      sizeof(params->points_uv[i]));
+            scaling_changed |= data->num_points_uv[i] !=
+                               obj->data.num_points_uv[i];
+            scaling_changed |= memcmp(data->points_uv[i],
+                                      obj->data.points_uv[i],
+                                      sizeof(data->points_uv[i]));
         }
     }
 
@@ -704,28 +746,28 @@ void pl_shader_av1_grain(struct pl_shader *sh,
         // SSBO layout in this case
         struct pl_var grain_y = pl_var_float("grain_y");
         grain_y.dim_a = GRAIN_WIDTH_LUT * GRAIN_HEIGHT_LUT;
-        ok &= sh_buf_desc_append(obj->tmp, sh->gpu, &obj->desc,
+        ok &= sh_buf_desc_append(obj->tmp, SH_GPU(sh), &obj->desc,
                                  &obj->layout_y, grain_y);
 
         struct pl_var grain_cb = pl_var_float("grain_cb");
         grain_cb.dim_a = chroma_lut_size;
-        ok &= sh_buf_desc_append(obj->tmp, sh->gpu, &obj->desc,
+        ok &= sh_buf_desc_append(obj->tmp, SH_GPU(sh), &obj->desc,
                                  &obj->layout_cb, grain_cb);
 
         struct pl_var grain_cr = pl_var_float("grain_cr");
         grain_cr.dim_a = chroma_lut_size;
-        ok &= sh_buf_desc_append(obj->tmp, sh->gpu, &obj->desc,
+        ok &= sh_buf_desc_append(obj->tmp, SH_GPU(sh), &obj->desc,
                                  &obj->layout_cr, grain_cr);
 
         struct pl_var offsets = pl_var_uint("offsets");
         offsets.dim_a = offsets_x * offsets_y;
-        ok &= sh_buf_desc_append(obj->tmp, sh->gpu, &obj->desc,
+        ok &= sh_buf_desc_append(obj->tmp, SH_GPU(sh), &obj->desc,
                                  &obj->layout_off, offsets);
 
         if (!ok) {
             PL_ERR(sh, "Failed generating SSBO buffer placement: Either GPU "
                    "limits exceeded or width/height nonsensical?");
-            return;
+            return false;
         }
 
         obj->num_offsets = offsets_x * offsets_y;
@@ -751,10 +793,10 @@ void pl_shader_av1_grain(struct pl_shader *sh,
         if (last_buf && last_buf->params.size > ssbo_params.size)
             ssbo_params.size = last_buf->params.size;
 
-        ssbo = pl_buf_pool_get(sh->gpu, &obj->ssbos, &ssbo_params);
+        ssbo = pl_buf_pool_get(SH_GPU(sh), &obj->ssbos, &ssbo_params);
         if (!ssbo) {
             PL_ERR(sh, "Failed creating/getting SSBO buffer for AV1 grain!");
-            return;
+            return false;
         }
 
         obj->desc.object = ssbo;
@@ -766,7 +808,7 @@ void pl_shader_av1_grain(struct pl_shader *sh,
 
         if (has_luma) {
             pl_assert(obj->layout_y.stride == sizeof(float));
-            pl_buf_write(sh->gpu, ssbo, obj->layout_y.offset, obj->grain,
+            pl_buf_write(SH_GPU(sh), ssbo, obj->layout_y.offset, obj->grain,
                          sizeof(obj->grain));
         }
 
@@ -774,22 +816,24 @@ void pl_shader_av1_grain(struct pl_shader *sh,
             generate_grain_uv(&obj->grain[0][0], obj->grain_tmp_uv,
                               obj->grain_tmp_y, PL_CHANNEL_CB, params);
             pl_assert(obj->layout_cb.stride == sizeof(float));
-            pl_buf_write(sh->gpu, ssbo, obj->layout_cb.offset, obj->grain,
+            pl_buf_write(SH_GPU(sh), ssbo, obj->layout_cb.offset, obj->grain,
                          sizeof(float) * chroma_lut_size);
 
             generate_grain_uv(&obj->grain[0][0], obj->grain_tmp_uv,
                               obj->grain_tmp_y, PL_CHANNEL_CR, params);
             pl_assert(obj->layout_cr.stride == sizeof(float));
-            pl_buf_write(sh->gpu, ssbo, obj->layout_cr.offset, obj->grain,
+            pl_buf_write(SH_GPU(sh), ssbo, obj->layout_cr.offset, obj->grain,
                          sizeof(float) * chroma_lut_size);
         }
 
-        generate_offsets(obj->offsets, offsets_x, offsets_y, params);
+        generate_offsets(obj->offsets, offsets_x, offsets_y, data);
         pl_assert(obj->layout_off.stride == sizeof(uint32_t));
-        pl_buf_write(sh->gpu, ssbo, obj->layout_off.offset, obj->offsets,
+        pl_buf_write(SH_GPU(sh), ssbo, obj->layout_off.offset, obj->offsets,
                      obj->num_offsets * obj->layout_off.stride);
 
-        obj->params = *params;
+        obj->data = *data;
+        obj->sub_x = params->sub_x;
+        obj->sub_y = params->sub_y;
     }
 
     // Update the scaling LUTs
@@ -798,16 +842,16 @@ void pl_shader_av1_grain(struct pl_shader *sh,
         struct {
             int num;
             const uint8_t (*points)[2];
-            const struct pl_grain_params *params;
+            const struct pl_av1_grain_data *data;
         } priv;
 
-        priv.params = params;
-        if (i == 0 || params->chroma_scaling_from_luma) {
-            priv.num = params->num_points_y;
-            priv.points = &params->points_y[0];
+        priv.data = data;
+        if (i == 0 || data->chroma_scaling_from_luma) {
+            priv.num = data->num_points_y;
+            priv.points = &data->points_y[0];
         } else {
-            priv.num = params->num_points_uv[i - 1];
-            priv.points = &params->points_uv[i - 1][0];
+            priv.num = data->num_points_uv[i - 1];
+            priv.points = &data->points_uv[i - 1][0];
         }
 
         if (priv.num > 0) {
@@ -817,7 +861,7 @@ void pl_shader_av1_grain(struct pl_shader *sh,
 
             if (!scaling[i]) {
                 PL_ERR(sh, "Failed generating/uploading scaling LUTs!");
-                return;
+                return false;
             }
         }
     }
@@ -844,7 +888,7 @@ void pl_shader_av1_grain(struct pl_shader *sh,
     }
 
     // Load the data vector which holds the offsets
-    GLSL("uint data = offsets[block_id.y * %d + block_id.x]; \n", offsets_x);
+    GLSL("uint data = offsets[block_id.y * %du + block_id.x]; \n", offsets_x);
 
     // If we need access to the external luma plane, load it now
     if (is_chroma) {
@@ -852,7 +896,7 @@ void pl_shader_av1_grain(struct pl_shader *sh,
         if (is_luma) {
             // We already have the luma channel as part of the pre-sampled color
             for (int i = 0; i < 3; i++) {
-                if (channels[i] == PL_CHANNEL_Y) {
+                if (params->channels[i] == PL_CHANNEL_Y) {
                     GLSL("averageLuma = color[%d]; \n", i);
                     break;
                 }
@@ -864,10 +908,10 @@ void pl_shader_av1_grain(struct pl_shader *sh,
                     .name = "luma",
                     .type = PL_DESC_SAMPLED_TEX,
                 },
-                .object = luma_tex,
+                .object = params->luma_tex,
             });
 
-            GLSL("pos = local_id * uvec2(%d, %d0;                \n"
+            GLSL("pos = local_id * uvec2(%d, %d0);               \n"
                  "averageLuma = texelFetch(%s, ivec2(pos), 0).r; \n",
                  1 << params->sub_x, 1 << params->sub_y,
                  luma);
@@ -875,9 +919,11 @@ void pl_shader_av1_grain(struct pl_shader *sh,
     }
 
     struct grain_scale scale = get_grain_scale(params);
-    int bits = params->repr.bits.color_depth;
+    int bits = PL_DEF(params->repr.bits.color_depth, 8);
+    pl_assert(bits >= 8);
+
     float minValue, maxLuma, maxChroma;
-    if (params->repr.levels == PL_COLOR_LEVELS_TV) {
+    if (pl_color_levels_guess(&params->repr) == PL_COLOR_LEVELS_TV) {
         float out_scale = (1 << bits) / ((1 << bits) - 1.0);
         out_scale /= scale.texture_scale;
         minValue  = 16  / 256.0 * out_scale;
@@ -892,7 +938,7 @@ void pl_shader_av1_grain(struct pl_shader *sh,
     }
 
     for (int i = 0; i < 3; i++) {
-        enum pl_channel c = channels[i];
+        enum pl_channel c = params->channels[i];
         if (c == PL_CHANNEL_NONE)
             continue;
         if (!scaling[c])
@@ -907,10 +953,10 @@ void pl_shader_av1_grain(struct pl_shader *sh,
                  i, i, minValue, maxLuma);
         } else {
             GLSL("val = averageLuma; \n");
-            if (!params->chroma_scaling_from_luma) {
+            if (!data->chroma_scaling_from_luma) {
                 GLSL("val = dot(vec2(val, color[%d]), vec2(%d.0, %d.0)); \n",
-                     i, params->uv_mult[c - 1], params->uv_mult[c - 1]);
-                int offset = params->uv_offset[c - 1] << (bits - 8);
+                     i, data->uv_mult[c - 1], data->uv_mult[c - 1]);
+                int offset = data->uv_offset[c - 1] << (bits - 8);
                 GLSL("val = val * 1.0/64.0 + %f; \n",
                      offset / scale.texture_scale);
             }
@@ -922,4 +968,5 @@ void pl_shader_av1_grain(struct pl_shader *sh,
     }
 
     GLSL("} \n");
+    return true;
 }
