@@ -39,7 +39,7 @@ void pl_shader_decode_color(struct pl_shader *sh, struct pl_color_repr *repr,
     // XYZ needs special handling due to the input gamma logic
     if (repr->sys == PL_COLOR_SYSTEM_XYZ) {
         float scale = pl_color_repr_normalize(repr);
-        GLSL("color.rgb = pow(%f * color.rgb, vec3(2.6));\n", scale);
+        GLSL("color.rgb = pow(vec3(%f) * color.rgb, vec3(2.6));\n", scale);
     }
 
     enum pl_color_system orig_sys = repr->sys;
@@ -69,11 +69,11 @@ void pl_shader_decode_color(struct pl_shader *sh, struct pl_color_repr *repr,
         // as per the BT.2020 specification, table 4. This is a non-linear
         // transformation because (constant) luminance receives non-equal
         // contributions from the three different channels.
-        GLSL("// constant luminance conversion                                \n"
-             "color.br = color.br * mix(vec2(1.5816, 0.9936),                 \n"
-             "                          vec2(1.9404, 1.7184),                 \n"
-             "                          %s(lessThanEqual(color.br, vec2(0)))) \n"
-             "           + color.gg;                                          \n",
+        GLSL("// constant luminance conversion                                  \n"
+             "color.br = color.br * mix(vec2(1.5816, 0.9936),                   \n"
+             "                          vec2(1.9404, 1.7184),                   \n"
+             "                          %s(lessThanEqual(color.br, vec2(0.0)))) \n"
+             "           + color.gg;                                            \n",
              sh_bvec(sh, 2));
         // Expand channels to camera-linear light. This shader currently just
         // assumes everything uses the BT.2020 12-bit gamma function, since the
@@ -86,10 +86,10 @@ void pl_shader_decode_color(struct pl_shader *sh, struct pl_color_repr *repr,
              sh_bvec(sh, 3));
         // Calculate the green channel from the expanded RYcB, and recompress to G'
         // The BT.2020 specification says Yc = 0.2627*R + 0.6780*G + 0.0593*B
-        GLSL("color.g = (lin.g - 0.2627*lin.r - 0.0593*lin.b)*1.0/0.6780; \n"
-             "color.g = mix(color.g * 4.5,                        \n"
-             "              1.0993 * pow(color.g, 0.45) - 0.0993, \n"
-             "              %s(0.0181 <= color.g));               \n",
+        GLSL("color.g = (lin.g - 0.2627*lin.r - 0.0593*lin.b)*1.0/0.6780;   \n"
+             "color.g = mix(color.g * 4.5,                                  \n"
+             "              1.0993 * pow(color.g, 0.45) - 0.0993,           \n"
+             "              %s(0.0181 <= color.g));                         \n",
              sh_bvec(sh, 1));
         break;
 
@@ -134,6 +134,17 @@ void pl_shader_decode_color(struct pl_shader *sh, struct pl_color_repr *repr,
         repr->alpha = PL_ALPHA_PREMULTIPLIED;
     }
 
+    // Gamma adjustment. Doing this here (in non-linear light) is technically
+    // somewhat wrong, but this is just an aesthetic parameter and not really
+    // meant for colorimetric precision, so we don't care too much.
+    if (params && params->gamma != 1.0) {
+        ident_t gamma = sh_var(sh, (struct pl_shader_var) {
+            .var = pl_var_float("gamma"),
+            .data = &params->gamma,
+        });
+        GLSL("color.rgb = pow(color.rgb, vec3(%s)); \n", gamma);
+    }
+
     GLSL("}\n");
 }
 
@@ -156,13 +167,61 @@ void pl_shader_encode_color(struct pl_shader *sh,
     GLSL("// pl_shader_encode_color \n"
          "{ \n");
 
-    if (!pl_color_system_is_linear(repr->sys)) {
-        // FIXME: implement this case
-        SH_FAIL(sh, "Non-linear color encoding currently unimplemented!");
-        return;
+    switch (repr->sys) {
+    case PL_COLOR_SYSTEM_BT_2020_C:
+        // Expand R'G'B' to RGB
+        GLSL("vec3 lin = mix(color.rgb * vec3(1.0/4.5),                        \n"
+             "                pow((color.rgb + vec3(0.0993))*vec3(1.0/1.0993), \n"
+             "                    vec3(1.0/0.45)),                             \n"
+             "                %s(lessThanEqual(vec3(0.08145), color.rgb)));    \n",
+             sh_bvec(sh, 3));
+
+        // Compute Yc from RGB and compress to R'Y'cB'
+        GLSL("color.g = dot(vec3(0.2627, 0.6780, 0.0593), lin);     \n"
+             "color.g = mix(color.g * 4.5,                          \n"
+             "              1.0993 * pow(color.g, 0.45) - 0.0993,   \n"
+             "              %s(0.0181 <= color.g));                 \n",
+             sh_bvec(sh, 1));
+
+        // Compute C'bc and C'rc into color.br
+        GLSL("color.br = color.br - color.gg;                           \n"
+             "color.br *= mix(vec2(1.0/1.5816, 1.0/0.9936),             \n"
+             "                vec2(1.0/1.9404, 1.0/1.7184),             \n"
+             "                %s(lessThanEqual(color.br, vec2(0.0))));  \n",
+             sh_bvec(sh, 2));
+        break;
+
+    case PL_COLOR_SYSTEM_BT_2100_PQ:
+    case PL_COLOR_SYSTEM_BT_2100_HLG: {
+        enum pl_color_transfer trc = repr->sys == PL_COLOR_SYSTEM_BT_2100_PQ
+                                        ? PL_COLOR_TRC_PQ
+                                        : PL_COLOR_TRC_HLG;
+
+        // Inverse of the matrix above
+        static const char *bt2100_rgb2lms = "mat3("
+            "0.412109, 0.166748, 0.024170, "
+            "0.523925, 0.720459, 0.075440, "
+            "0.063965, 0.112793, 0.900394) ";
+
+        pl_shader_linearize(sh, trc);
+        GLSL("color.rgb = %s * color.rgb; \n", bt2100_rgb2lms);
+        pl_shader_delinearize(sh, trc);
+        break;
+    }
+
+    case PL_COLOR_SYSTEM_XYZ:
+        break; // no special pre-processing needed
+
+    default:
+        assert(pl_color_system_is_linear(repr->sys));
+        break;
     }
 
     struct pl_color_repr copy = *repr;
+    float xyzscale = (repr->sys == PL_COLOR_SYSTEM_XYZ)
+                        ? pl_color_repr_normalize(&copy)
+                        : 0.0;
+
     struct pl_transform3x3 tr = pl_color_repr_decode(&copy, NULL);
     pl_transform3x3_invert(&tr);
 
@@ -177,6 +236,8 @@ void pl_shader_encode_color(struct pl_shader *sh,
     });
 
     GLSL("color.rgb = %s * color.rgb + %s;\n", cmat, cmat_c);
+    if (repr->sys == PL_COLOR_SYSTEM_XYZ)
+        GLSL("color.rgb = pow(color.rgb, vec3(1.0/2.6)) * vec3(1.0/%f); \n", xyzscale);
     if (repr->alpha == PL_ALPHA_INDEPENDENT)
         GLSL("color.rgb /= vec3(max(color.a, 1e-6));\n");
     GLSL("}\n");
@@ -539,6 +600,8 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
                            struct pl_shader_obj **state,
                            const struct pl_peak_detect_params *params)
 {
+    const struct pl_gpu *gpu = SH_GPU(sh);
+
     params = PL_DEF(params, &pl_peak_detect_default_params);
     if (!sh_require(sh, PL_SHADER_SIG_COLOR, 0, 0))
         return false;
@@ -548,13 +611,18 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
         return false;
     }
 
+    if (gpu->glsl.version < 130) {
+        // uint was added in GLSL 130
+        PL_ERR(sh, "HDR peak detection requires GLSL >= 130!");
+        return false;
+    }
+
     struct sh_peak_obj *obj;
     obj = SH_OBJ(sh, state, PL_SHADER_OBJ_PEAK_DETECT, struct sh_peak_obj,
                  sh_peak_uninit);
     if (!obj)
         return false;
 
-    const struct pl_gpu *gpu = SH_GPU(sh);
     obj->gpu = gpu;
 
     if (!obj->buf) {
@@ -633,7 +701,7 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
     // Have one thread per work group update the global atomics. Do this
     // at the end of the shader to avoid clobbering `average`, in case the
     // state object will be used by the same pass.
-    GLSLF("if (gl_LocalInvocationIndex == 0) {                                  \n"
+    GLSLF("if (gl_LocalInvocationIndex == 0u) {                                 \n"
           "    int wg_avg = %s / int(gl_WorkGroupSize.x * gl_WorkGroupSize.y);  \n"
           "    atomicAdd(frame_sum, wg_avg);                                    \n"
           "    atomicMax(frame_max, %s);                                        \n"
@@ -643,7 +711,7 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
 
     // Finally, to update the global state per dispatch, we increment a counter
     GLSLF("    uint num_wg = gl_NumWorkGroups.x * gl_NumWorkGroups.y;           \n"
-          "    if (atomicAdd(counter, 1) == num_wg - 1) {                       \n"
+          "    if (atomicAdd(counter, 1u) == num_wg - 1u) {                     \n"
           "        vec2 cur = vec2(float(frame_sum) / float(num_wg), frame_max);\n"
           "        cur *= vec2(1.0 / %f, 1.0 / %f);                             \n"
           "        cur.x = exp(cur.x);                                          \n",
@@ -669,7 +737,7 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
     // Reset SSBO state for the next frame
     GLSLF("        frame_sum = 0;            \n"
           "        frame_max = 0;            \n"
-          "        counter = 0;              \n"
+          "        counter = 0u;             \n"
           "        memoryBarrierBuffer();    \n"
           "    }                             \n"
           "}                                 \n");
@@ -1115,7 +1183,7 @@ done: ;
 
     default: // LUT-based methods
         pl_assert(lut);
-        GLSL("bias = %s(ivec2(pos * %d));\n", lut, lut_size);
+        GLSL("bias = %s(ivec2(pos * %d.0));\n", lut, lut_size);
         break;
     }
 
