@@ -14,7 +14,7 @@ int main()
     }
 
     float pq_peak = pl_color_transfer_nominal_peak(PL_COLOR_TRC_PQ);
-    REQUIRE(feq(PL_COLOR_REF_WHITE * pq_peak, 10000));
+    REQUIRE(feq(PL_COLOR_REF_WHITE * pq_peak, 10000, 1e-7));
 
     struct pl_color_repr tv_repr = {
         .sys       = PL_COLOR_SYSTEM_BT_709,
@@ -30,8 +30,8 @@ int main()
     for (int bits = 1; bits <= 16; bits++) {
         tv_repr.bits.color_depth = tv_repr.bits.sample_depth = bits;
         pc_repr.bits.color_depth = pc_repr.bits.sample_depth = bits;
-        REQUIRE(feq(pl_color_repr_normalize(&tv_repr), 1.0));
-        REQUIRE(feq(pl_color_repr_normalize(&pc_repr), 1.0));
+        REQUIRE(feq(pl_color_repr_normalize(&tv_repr), 1.0, 1e-7));
+        REQUIRE(feq(pl_color_repr_normalize(&pc_repr), 1.0, 1e-7));
     }
 
     tv_repr.bits.color_depth  = 8;
@@ -43,15 +43,15 @@ int main()
     float tv8to12 = pl_color_repr_normalize(&tv_repr);
 
     // Simulate the effect of GPU texture sampling on UNORM texture
-    REQUIRE(feq(tv8to10 * 16 /1023.,  64/1023.)); // black
-    REQUIRE(feq(tv8to10 * 235/1023., 940/1023.)); // nominal white
-    REQUIRE(feq(tv8to10 * 128/1023., 512/1023.)); // achromatic
-    REQUIRE(feq(tv8to10 * 240/1023., 960/1023.)); // nominal chroma peak
+    REQUIRE(feq(tv8to10 * 16 /1023.,  64/1023., 1e-7)); // black
+    REQUIRE(feq(tv8to10 * 235/1023., 940/1023., 1e-7)); // nominal white
+    REQUIRE(feq(tv8to10 * 128/1023., 512/1023., 1e-7)); // achromatic
+    REQUIRE(feq(tv8to10 * 240/1023., 960/1023., 1e-7)); // nominal chroma peak
 
-    REQUIRE(feq(tv8to12 * 16 /4095., 256 /4095.)); // black
-    REQUIRE(feq(tv8to12 * 235/4095., 3760/4095.)); // nominal white
-    REQUIRE(feq(tv8to12 * 128/4095., 2048/4095.)); // achromatic
-    REQUIRE(feq(tv8to12 * 240/4095., 3840/4095.)); // nominal chroma peak
+    REQUIRE(feq(tv8to12 * 16 /4095., 256 /4095., 1e-7)); // black
+    REQUIRE(feq(tv8to12 * 235/4095., 3760/4095., 1e-7)); // nominal white
+    REQUIRE(feq(tv8to12 * 128/4095., 2048/4095., 1e-7)); // achromatic
+    REQUIRE(feq(tv8to12 * 240/4095., 3840/4095., 1e-7)); // nominal chroma peak
 
     // Ensure lavc's xyz12 is handled correctly
     struct pl_color_repr xyz12 = {
@@ -65,7 +65,7 @@ int main()
     };
 
     float xyz = pl_color_repr_normalize(&xyz12);
-    REQUIRE(feq(xyz * (4095 << 4), 65535));
+    REQUIRE(feq(xyz * (4095 << 4), 65535, 1e-7));
 
     // Assume we uploaded a 10-bit source directly (unshifted) as a 16-bit
     // texture. This texture multiplication factor should make it behave as if
@@ -73,7 +73,7 @@ int main()
     pc_repr.bits.color_depth = 10;
     pc_repr.bits.sample_depth = 16;
     float pc10to16 = pl_color_repr_normalize(&pc_repr);
-    REQUIRE(feq(pc10to16 * 1000/65535., 1000/1023.));
+    REQUIRE(feq(pc10to16 * 1000/65535., 1000/1023., 1e-7));
 
     const struct pl_raw_primaries *bt709, *bt2020;
     bt709 = pl_raw_primaries_get(PL_COLOR_PRIM_BT_709);
@@ -87,20 +87,20 @@ int main()
     // Make sure the double-inversion round trips
     for (int y = 0; y < 3; y++) {
         for (int x = 0; x < 3; x++)
-            REQUIRE(feq(rgb2xyz.m[y][x], rgb2xyz_.m[y][x]));
+            REQUIRE(feq(rgb2xyz.m[y][x], rgb2xyz_.m[y][x], 1e-6));
     }
 
     // Make sure mapping the spectral RGB colors (i.e. the matrix rows) matches
     // our original primaries
     float Y = rgb2xyz.m[1][0];
-    REQUIRE(feq(rgb2xyz.m[0][0], pl_cie_X(bt709->red) * Y));
-    REQUIRE(feq(rgb2xyz.m[2][0], pl_cie_Z(bt709->red) * Y));
+    REQUIRE(feq(rgb2xyz.m[0][0], pl_cie_X(bt709->red) * Y, 1e-7));
+    REQUIRE(feq(rgb2xyz.m[2][0], pl_cie_Z(bt709->red) * Y, 1e-7));
     Y = rgb2xyz.m[1][1];
-    REQUIRE(feq(rgb2xyz.m[0][1], pl_cie_X(bt709->green) * Y));
-    REQUIRE(feq(rgb2xyz.m[2][1], pl_cie_Z(bt709->green) * Y));
+    REQUIRE(feq(rgb2xyz.m[0][1], pl_cie_X(bt709->green) * Y, 1e-7));
+    REQUIRE(feq(rgb2xyz.m[2][1], pl_cie_Z(bt709->green) * Y, 1e-7));
     Y = rgb2xyz.m[1][2];
-    REQUIRE(feq(rgb2xyz.m[0][2], pl_cie_X(bt709->blue) * Y));
-    REQUIRE(feq(rgb2xyz.m[2][2], pl_cie_Z(bt709->blue) * Y));
+    REQUIRE(feq(rgb2xyz.m[0][2], pl_cie_X(bt709->blue) * Y, 1e-7));
+    REQUIRE(feq(rgb2xyz.m[2][2], pl_cie_Z(bt709->blue) * Y, 1e-7));
 
     // Make sure the gamut mapping round-trips
     struct pl_matrix3x3 bt709_bt2020, bt2020_bt709;
@@ -112,7 +112,7 @@ int main()
         pl_matrix3x3_apply(&bt709_bt2020, dst);
         pl_matrix3x3_apply(&bt2020_bt709, dst);
         for (int i = 0; i < 3; i++)
-            REQUIRE(feq(dst[i], vec[i]));
+            REQUIRE(feq(dst[i], vec[i], 1e-6));
     }
 
     // Ensure the decoding matrix round-trips to white/black
@@ -144,14 +144,14 @@ int main()
         }
 
         pl_transform3x3_apply(&yuv2rgb, white);
-        REQUIRE(feq(white[0], 1.0));
-        REQUIRE(feq(white[1], 1.0));
-        REQUIRE(feq(white[2], 1.0));
+        REQUIRE(feq(white[0], 1.0, 1e-6));
+        REQUIRE(feq(white[1], 1.0, 1e-6));
+        REQUIRE(feq(white[2], 1.0, 1e-6));
 
         pl_transform3x3_apply(&yuv2rgb, black);
-        REQUIRE(feq(black[0], 0.0));
-        REQUIRE(feq(black[1], 0.0));
-        REQUIRE(feq(black[2], 0.0));
+        REQUIRE(feq(black[0], 0.0, 1e-6));
+        REQUIRE(feq(black[1], 0.0, 1e-6));
+        REQUIRE(feq(black[2], 0.0, 1e-6));
     }
 
     // Make sure chromatic adaptation works
@@ -164,7 +164,7 @@ int main()
 
     float white[3] = { 1.0, 1.0, 1.0 };
     pl_matrix3x3_apply(&d50_d65, white);
-    REQUIRE(feq(white[0], 1.0) && feq(white[1], 1.0) && feq(white[2], 1.0));
+    REQUIRE(feq(white[0], 1.0, 1e-6) && feq(white[1], 1.0, 1e-6) && feq(white[2], 1.0, 1e-6));
 
     // Simulate a typical 10-bit YCbCr -> 16 bit texture conversion
     tv_repr.bits.color_depth  = 10;
@@ -173,9 +173,9 @@ int main()
     yuv2rgb = pl_color_repr_decode(&tv_repr, NULL);
     float test[3] = { 575/65535., 336/65535., 640/65535. };
     pl_transform3x3_apply(&yuv2rgb, test);
-    REQUIRE(feq(test[0], 0.808305));
-    REQUIRE(feq(test[1], 0.553254));
-    REQUIRE(feq(test[2], 0.218841));
+    REQUIRE(feq(test[0], 0.808305, 1e-6));
+    REQUIRE(feq(test[1], 0.553254, 1e-6));
+    REQUIRE(feq(test[2], 0.218841, 1e-6));
 
     // DVD
     REQUIRE(pl_color_system_guess_ycbcr(720, 480) == PL_COLOR_SYSTEM_BT_601);
