@@ -1,17 +1,79 @@
 #include "tests.h"
 #include "shaders.h"
 
-static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex,
-                              uint8_t *src, uint8_t *dst)
+static uint8_t test_src[16*16*16 * 4 * sizeof(double)] = {0};
+static uint8_t test_dst[16*16*16 * 4 * sizeof(double)] = {0};
+
+static void pl_buffer_tests(const struct pl_gpu *gpu)
 {
-    if (!tex)
+    const size_t buf_size = 1024;
+    assert(buf_size <= sizeof(test_src));
+
+    memset(test_dst, 0, buf_size);
+    for (int i = 0; i < buf_size; i++)
+        test_src[i] = (RANDOM * 256);
+
+    const struct pl_buf *buf = NULL;
+    if (buf_size > gpu->limits.max_xfer_size)
         return;
 
-    int texels = tex->params.w;
-    texels *= tex->params.h ? tex->params.h : 1;
-    texels *= tex->params.d ? tex->params.d : 1;
+    printf("test xfer buffer static creation and readback\n");
+    buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+        .type = PL_BUF_TEX_TRANSFER,
+        .size = buf_size,
+        .host_readable = true,
+        .initial_data = test_src,
+    });
 
-    const struct pl_fmt *fmt = tex->params.format;
+    REQUIRE(buf);
+    REQUIRE(pl_buf_read(gpu, buf, 0, test_dst, buf_size));
+    REQUIRE(memcmp(test_src, test_dst, buf_size) == 0);
+    pl_buf_destroy(gpu, &buf);
+
+    printf("test xfer buffer empty creation, update and readback\n");
+    memset(test_dst, 0, buf_size);
+    buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+        .type = PL_BUF_TEX_TRANSFER,
+        .size = buf_size,
+        .host_writable = true,
+        .host_readable = true,
+    });
+
+    REQUIRE(buf);
+    pl_buf_write(gpu, buf, 0, test_src, buf_size);
+    REQUIRE(pl_buf_read(gpu, buf, 0, test_dst, buf_size));
+    REQUIRE(memcmp(test_src, test_dst, buf_size) == 0);
+    pl_buf_destroy(gpu, &buf);
+
+    if (gpu->caps & PL_GPU_CAP_MAPPED_BUFFERS) {
+        printf("test host mapped buffer readback\n");
+        buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+            .type = PL_BUF_TEX_TRANSFER,
+            .size = buf_size,
+            .host_mapped = true,
+            .initial_data = test_src,
+        });
+
+        REQUIRE(buf);
+        REQUIRE(!pl_buf_poll(gpu, buf, 0));
+        REQUIRE(memcmp(test_src, buf->data, buf_size) == 0);
+        pl_buf_destroy(gpu, &buf);
+    }
+}
+
+static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex[2],
+                              uint8_t *src, uint8_t *dst)
+{
+    if (!tex[0] || !tex[1]) {
+        printf("failed creating test textures... skipping this test\n");
+        return;
+    }
+
+    int texels = tex[0]->params.w;
+    texels *= tex[0]->params.h ? tex[0]->params.h : 1;
+    texels *= tex[0]->params.d ? tex[0]->params.d : 1;
+
+    const struct pl_fmt *fmt = tex[0]->params.format;
     size_t bytes = texels * fmt->texel_size;
     memset(src, 0, bytes);
     memset(dst, 0, bytes);
@@ -20,12 +82,29 @@ static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex
         src[i] = (RANDOM * 256);
 
     REQUIRE(pl_tex_upload(gpu, &(struct pl_tex_transfer_params){
-        .tex = tex,
+        .tex = tex[0],
         .ptr = src,
     }));
 
+    // Test blitting, if possible for this format
+    const struct pl_tex *dst_tex = tex[0];
+    if (tex[0]->params.blit_src && tex[1]->params.blit_dst) {
+        struct pl_rect3d rc = {
+            .x0 = 0,
+            .y0 = 0,
+            .z0 = 0,
+            .x1 = tex[0]->params.w,
+            .y1 = tex[0]->params.h,
+            .z1 = tex[0]->params.d,
+        };
+
+        pl_tex_clear(gpu, tex[1], (float[4]){0.0}); // for testing
+        pl_tex_blit(gpu, tex[1], tex[0], rc, rc);
+        dst_tex = tex[1];
+    }
+
     REQUIRE(pl_tex_download(gpu, &(struct pl_tex_transfer_params){
-        .tex = tex,
+        .tex = dst_tex,
         .ptr = dst,
     }));
 
@@ -38,46 +117,61 @@ static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex
     }
 }
 
-static uint8_t test_src[16*16*16 * 4 * sizeof(double)] = {0};
-static uint8_t test_dst[16*16*16 * 4 * sizeof(double)] = {0};
-
 static void pl_texture_tests(const struct pl_gpu *gpu)
 {
-    for (int i = 0; i < gpu->num_formats; i++) {
-        const struct pl_fmt *fmt = gpu->formats[i];
-        if (fmt->opaque)
+    for (int f = 0; f < gpu->num_formats; f++) {
+        const struct pl_fmt *fmt = gpu->formats[f];
+        if (fmt->opaque || !(fmt->caps & PL_FMT_CAP_HOST_READABLE))
             continue;
 
         printf("testing texture roundtrip for format %s\n", fmt->name);
         assert(fmt->texel_size <= 4 * sizeof(double));
 
-        struct pl_tex_params params = {
+        struct pl_tex_params ref_params = {
             .format        = fmt,
+            .blit_src      = (fmt->caps & PL_FMT_CAP_BLITTABLE),
+            .blit_dst      = (fmt->caps & PL_FMT_CAP_BLITTABLE),
             .host_writable = true,
             .host_readable = true,
         };
 
-        const struct pl_tex *tex = NULL;
+        const struct pl_tex *tex[2];
 
         if (gpu->limits.max_tex_1d_dim >= 16) {
+            printf("... 1D\n");
+            struct pl_tex_params params = ref_params;
             params.w = 16;
-            tex = pl_tex_create(gpu, &params);
+            if (!(gpu->caps & PL_GPU_CAP_BLITTABLE_1D_3D))
+                params.blit_src = params.blit_dst = false;
+            for (int i = 0; i < PL_ARRAY_SIZE(tex); i++)
+                tex[i] = pl_tex_create(gpu, &params);
             pl_test_roundtrip(gpu, tex, test_src, test_dst);
-            pl_tex_destroy(gpu, &tex);
+            for (int i = 0; i < PL_ARRAY_SIZE(tex); i++)
+                pl_tex_destroy(gpu, &tex[i]);
         }
 
         if (gpu->limits.max_tex_2d_dim >= 16) {
+            printf("... 2D\n");
+            struct pl_tex_params params = ref_params;
             params.w = params.h = 16;
-            tex = pl_tex_create(gpu, &params);
+            for (int i = 0; i < PL_ARRAY_SIZE(tex); i++)
+                tex[i] = pl_tex_create(gpu, &params);
             pl_test_roundtrip(gpu, tex, test_src, test_dst);
-            pl_tex_destroy(gpu, &tex);
+            for (int i = 0; i < PL_ARRAY_SIZE(tex); i++)
+                pl_tex_destroy(gpu, &tex[i]);
         }
 
         if (gpu->limits.max_tex_3d_dim >= 16) {
+            printf("... 3D\n");
+            struct pl_tex_params params = ref_params;
             params.w = params.h = params.d = 16;
-            tex = pl_tex_create(gpu, &params);
+            if (!(gpu->caps & PL_GPU_CAP_BLITTABLE_1D_3D))
+                params.blit_src = params.blit_dst = false;
+            for (int i = 0; i < PL_ARRAY_SIZE(tex); i++)
+                tex[i] = pl_tex_create(gpu, &params);
             pl_test_roundtrip(gpu, tex, test_src, test_dst);
-            pl_tex_destroy(gpu, &tex);
+            for (int i = 0; i < PL_ARRAY_SIZE(tex); i++)
+                pl_tex_destroy(gpu, &tex[i]);
         }
     }
 }
@@ -173,28 +267,37 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
         .vertex_count   = sizeof(vertices) / sizeof(struct vertex),
     });
 
-    static float data[FBO_H * FBO_W * 4] = {0};
-    pl_tex_download(gpu, &(struct pl_tex_transfer_params) {
-        .tex = fbo,
-        .ptr = data,
-    });
-
-    for (int y = 0; y < FBO_H; y++) {
-        for (int x = 0; x < FBO_W; x++) {
-            float *color = &data[(y * FBO_W + x) * 4];
-            printf("color: %f %f %f %f\n", color[0], color[1], color[2], color[3]);
-            REQUIRE(feq(color[0], (x + 0.5) / FBO_W));
-            REQUIRE(feq(color[1], (y + 0.5) / FBO_H));
-            REQUIRE(feq(color[2], 0.0));
-            REQUIRE(feq(color[3], 1.0));
-        }
-    }
-
     pl_pass_destroy(gpu, &pass);
-    pl_tex_clear(gpu, fbo, (float[4]){0});
+
+    static float data[FBO_H * FBO_W * 4] = {0};
+
+    // Test against the known pattern of `src`, only useful for roundtrip tests
+#define TEST_FBO_PATTERN(eps, fmt, ...)                                     \
+    do {                                                                    \
+        printf("testing pattern of " fmt "\n", __VA_ARGS__);                \
+        REQUIRE(pl_tex_download(gpu, &(struct pl_tex_transfer_params) {     \
+            .tex = fbo,                                                     \
+            .ptr = data,                                                    \
+        }));                                                                \
+                                                                            \
+        for (int y = 0; y < FBO_H; y++) {                                   \
+            for (int x = 0; x < FBO_W; x++) {                               \
+                float *color = &data[(y * FBO_W + x) * 4];                  \
+                printf("color[%d][%d] = %f %f %f %f\n",                     \
+                       y, x, color[0], color[1], color[2], color[3]);       \
+                REQUIRE(feq(color[0], (x + 0.5) / FBO_W, eps));             \
+                REQUIRE(feq(color[1], (y + 0.5) / FBO_H, eps));             \
+                REQUIRE(feq(color[2], 0.0, eps));                           \
+                REQUIRE(feq(color[3], 1.0, eps));                           \
+            }                                                               \
+        }                                                                   \
+    } while (0)
+
+    TEST_FBO_PATTERN(1e-6, "%s", "initial rendering");
 
     // Test the use of pl_dispatch
     struct pl_dispatch *dp = pl_dispatch_create(gpu->ctx, gpu);
+    struct pl_shader *sh;
 
     const struct pl_tex *src;
     src = pl_tex_create(gpu, &(struct pl_tex_params) {
@@ -206,10 +309,48 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
         .initial_data   = data,
     });
 
+    // Test encoding/decoding of all gamma functions, color spaces, etc.
+    for (enum pl_color_transfer trc = 0; trc < PL_COLOR_TRC_COUNT; trc++) {
+        sh = pl_dispatch_begin(dp);
+        pl_shader_sample_direct(sh, &(struct pl_sample_src) { .tex = src });
+        pl_shader_delinearize(sh, trc);
+        pl_shader_linearize(sh, trc);
+        REQUIRE(pl_dispatch_finish(dp, &sh, fbo, NULL, NULL));
+
+        float epsilon = pl_color_transfer_is_hdr(trc) ? 1e-4 : 1e-6;
+        TEST_FBO_PATTERN(epsilon, "transfer function %d", (int) trc);
+    }
+
+    for (enum pl_color_system sys = 0; sys < PL_COLOR_SYSTEM_COUNT; sys++) {
+        sh = pl_dispatch_begin(dp);
+        pl_shader_sample_direct(sh, &(struct pl_sample_src) { .tex = src });
+        pl_shader_encode_color(sh, &(struct pl_color_repr) { .sys = sys });
+        pl_shader_decode_color(sh, &(struct pl_color_repr) { .sys = sys }, NULL);
+        REQUIRE(pl_dispatch_finish(dp, &sh, fbo, NULL, NULL));
+
+        float epsilon;
+        switch (sys) {
+        case PL_COLOR_SYSTEM_BT_2020_C:
+            epsilon = 1e-5;
+            break;
+
+        case PL_COLOR_SYSTEM_BT_2100_PQ:
+        case PL_COLOR_SYSTEM_BT_2100_HLG:
+            // These seem to be horrifically noisy and prone to breaking on
+            // edge cases for some reason
+            // TODO: figure out why!
+            continue;
+            break;
+
+        default: epsilon = 1e-6; break;
+        }
+
+        TEST_FBO_PATTERN(epsilon, "color system %d", (int) sys);
+    }
+
     // Repeat this a few times to test the caching
     for (int i = 0; i < 10; i++) {
-        printf("iteration %d\n", i);
-        struct pl_shader *sh = pl_dispatch_begin(dp);
+        sh = pl_dispatch_begin(dp);
 
         // For testing, force the use of CS if possible
         if (gpu->caps & PL_GPU_CAP_COMPUTE) {
@@ -227,28 +368,10 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
                 .grain          = 0.0,
         });
 
-        pl_shader_linearize(sh, PL_COLOR_TRC_GAMMA22);
         REQUIRE(pl_dispatch_finish(dp, &sh, fbo, NULL, NULL));
+        TEST_FBO_PATTERN(1e-6, "deband iter %d", i);
     }
 
-    pl_tex_download(gpu, &(struct pl_tex_transfer_params) {
-        .tex = fbo,
-        .ptr = data,
-    });
-
-    for (int y = 0; y < FBO_H; y++) {
-        for (int x = 0; x < FBO_W; x++) {
-            float *color = &data[(y * FBO_W + x) * 4];
-            printf("color: %f %f %f %f\n", color[0], color[1], color[2], color[3]);
-            REQUIRE(feq(color[0], pow((x + 0.5) / FBO_W, 2.2)));
-            REQUIRE(feq(color[1], pow((y + 0.5) / FBO_H, 2.2)));
-            REQUIRE(feq(color[2], 0.0));
-            REQUIRE(feq(color[3], 1.0));
-        }
-    }
-
-    struct pl_shader *sh;
-
 #if PL_HAVE_LCMS
     // Test the use of 3DLUTs if available
     sh = pl_dispatch_begin(dp);
@@ -271,15 +394,22 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
     // Test AV1 grain synthesis
     struct pl_shader_obj *grain = NULL;
     for (int i = 0; i < 2; i++) {
-        struct pl_grain_params grain_params = av1_grain_params;
-        grain_params.width = FBO_W;
-        grain_params.height = FBO_H;
-        grain_params.grain_seed = rand();
-        grain_params.overlap = !!i;
+        struct pl_av1_grain_params grain_params = {
+            .data = av1_grain_data,
+            .luma_tex = src,
+            .channels = { 0, 1, 2 },
+            .repr = {
+                .sys = PL_COLOR_SYSTEM_BT_709,
+                .levels = PL_COLOR_LEVELS_TV,
+                .bits = { .color_depth = 10, .sample_depth = 10 },
+            },
+        };
+        grain_params.data.grain_seed = rand();
+        grain_params.data.overlap = !!i;
+
         sh = pl_dispatch_begin(dp);
         pl_shader_sample_direct(sh, &(struct pl_sample_src) { .tex = src });
-        pl_shader_av1_grain(sh, &grain, (enum pl_channel[]){0, 1, 2}, NULL,
-                            &grain_params);
+        pl_shader_av1_grain(sh, &grain, &grain_params);
         REQUIRE(pl_dispatch_finish(dp, &sh, fbo, NULL, NULL));
     }
     pl_shader_obj_destroy(&grain);
@@ -320,14 +450,21 @@ static void pl_scaler_tests(const struct pl_gpu *gpu)
         .initial_data   = &data_5x5[0][0],
     });
 
-    const struct pl_tex *fbo = pl_tex_create(gpu, &(struct pl_tex_params) {
+    struct pl_tex_params fbo_params = {
         .w              = 100,
         .h              = 100,
         .format         = fbo_fmt,
         .renderable     = true,
         .storable       = !!(fbo_fmt->caps & PL_FMT_CAP_STORABLE),
         .host_readable  = true,
-    });
+    };
+
+    const struct pl_tex *fbo = pl_tex_create(gpu, &fbo_params);
+    if (!fbo) {
+        printf("Failed creating readable FBO... falling back to non-readable\n");
+        fbo_params.host_readable = false;
+        fbo = pl_tex_create(gpu, &fbo_params);
+    }
 
     struct pl_dispatch *dp = pl_dispatch_create(gpu->ctx, gpu);
     if (!dot5x5 || !fbo || !dp)
@@ -348,20 +485,22 @@ static void pl_scaler_tests(const struct pl_gpu *gpu)
     ));
     REQUIRE(pl_dispatch_finish(dp, &sh, fbo, NULL, NULL));
 
-    fbo_data = malloc(fbo->params.w * fbo->params.h * sizeof(float));
-    REQUIRE(pl_tex_download(gpu, &(struct pl_tex_transfer_params) {
-        .tex            = fbo,
-        .ptr            = fbo_data,
-    }));
+    if (fbo->params.host_readable) {
+        fbo_data = malloc(fbo->params.w * fbo->params.h * sizeof(float));
+        REQUIRE(pl_tex_download(gpu, &(struct pl_tex_transfer_params) {
+            .tex            = fbo,
+            .ptr            = fbo_data,
+        }));
 
-    int max = 255;
-    printf("P2\n%d %d\n%d\n", fbo->params.w, fbo->params.h, max);
-    for (int y = 0; y < fbo->params.h; y++) {
-        for (int x = 0; x < fbo->params.w; x++) {
-            float v = fbo_data[y * fbo->params.h + x];
-            printf("%d ", (int) round(fmin(fmax(v, 0.0), 1.0) * max));
+        int max = 255;
+        printf("P2\n%d %d\n%d\n", fbo->params.w, fbo->params.h, max);
+        for (int y = 0; y < fbo->params.h; y++) {
+            for (int x = 0; x < fbo->params.w; x++) {
+                float v = fbo_data[y * fbo->params.h + x];
+                printf("%d ", (int) round(fmin(fmax(v, 0.0), 1.0) * max));
+            }
+            printf("\n");
         }
-        printf("\n");
     }
 
 error:
@@ -463,6 +602,7 @@ static void pl_render_tests(const struct pl_gpu *gpu)
     do {                                                                \
         for (int i = 0; i <= LIMIT; i++) {                              \
             struct pl_render_params params = pl_render_default_params;  \
+            params.force_dither = true;                                 \
             struct STYPE tmp = DEFAULT;                                 \
             tmp.FIELD = i;                                              \
             params.SNAME = &tmp;                                        \
@@ -488,6 +628,7 @@ static void pl_render_tests(const struct pl_gpu *gpu)
     TEST_PARAMS(color_map, intent, PL_INTENT_ABSOLUTE_COLORIMETRIC);
     TEST_PARAMS(color_map, gamut_warning, 1);
     TEST_PARAMS(dither, method, PL_DITHER_WHITE_NOISE);
+    TEST_PARAMS(dither, temporal, true);
     TEST(cone_params, pl_cone_params, pl_vision_deuteranomaly, strength, 0);
 
     // Test HDR stuff
@@ -505,8 +646,39 @@ static void pl_render_tests(const struct pl_gpu *gpu)
         .saturation = 1.5,
         .gamma = 0.8,
     };
-
     REQUIRE(pl_render_image(rr, &image, &target, &params));
+    params = pl_render_default_params;
+
+    params.force_3dlut = true;
+    REQUIRE(pl_render_image(rr, &image, &target, &params));
+    params = pl_render_default_params;
+
+    image.av1_grain = av1_grain_data;
+    REQUIRE(pl_render_image(rr, &image, &target, &params));
+    image.av1_grain = (struct pl_av1_grain_data) {0};
+
+    // Test overlays
+    image.num_overlays = 1;
+    image.overlays = &(struct pl_overlay) {
+        .plane = img5x5,
+        .rect = {0, 0, 2, 2},
+        .mode = PL_OVERLAY_NORMAL,
+    };
+    REQUIRE(pl_render_image(rr, &image, &target, &params));
+    params.disable_fbos = true;
+    REQUIRE(pl_render_image(rr, &image, &target, &params));
+    image.num_overlays = 0;
+    params = pl_render_default_params;
+
+    target.num_overlays = 1;
+    target.overlays = &(struct pl_overlay) {
+        .plane = img5x5,
+        .rect = {5, 5, 15, 15},
+        .mode = PL_OVERLAY_MONOCHROME,
+        .base_color = {1.0, 0.5, 0.0},
+    };
+    REQUIRE(pl_render_image(rr, &image, &target, &params));
+    target.num_overlays = 0;
 
 error:
     free(fbo_data);
@@ -517,6 +689,7 @@ error:
 
 static void gpu_tests(const struct pl_gpu *gpu)
 {
+    pl_buffer_tests(gpu);
     pl_texture_tests(gpu);
     pl_shader_tests(gpu);
     pl_scaler_tests(gpu);
